'''OpenGL extension NV.gpu_program5

This module customises the behaviour of the 
OpenGL.raw.GL.NV.gpu_program5 to provide a more 
Python-friendly API

Overview (from the spec)
	
	This specification documents the common instruction set and basic
	functionality provided by NVIDIA's 5th generation of assembly instruction
	sets supporting programmable graphics pipeline stages.
	
	The instruction set builds upon the basic framework provided by the
	ARB_vertex_program and ARB_fragment_program extensions to expose
	considerably more capable hardware.  In addition to new capabilities for
	vertex and fragment programs, this extension provides new functionality
	for geometry programs as originally described in the NV_geometry_program4
	specification, and serves as the basis for the new tessellation control
	and evaluation programs described in the NV_tessellation_program5
	extension.
	
	Programs using the functionality provided by this extension should begin
	with the program headers "!!NVvp5.0" (vertex programs), "!!NVtcp5.0"
	(tessellation control programs), "!!NVtep5.0" (tessellation evaluation
	programs), "!!NVgp5.0" (geometry programs), and "!!NVfp5.0" (fragment
	programs).
	
	This extension provides a variety of new features, including:
	
	  * support for 64-bit integer operations;
	
	  * the ability to dynamically index into an array of texture units or
	    program parameter buffers;
	
	  * extending texel offset support to allow loading texel offsets from
	    regular integer operands computed at run-time, instead of requiring
	    that the offsets be constants encoded in texture instructions;
	
	  * extending TXG (texture gather) support to return the 2x2 footprint
	    from any component of the texture image instead of always returning
	    the first (x) component;
	
	  * extending TXG to support shadow comparisons in conjunction with a
	    depth texture, via the SHADOW* targets;
	
	  * further extending texture gather support to provide a new opcode
	    (TXGO) that applies a separate texel offset vector to each of the four
	    samples returned by the instruction; 
	
	  * bit manipulation instructions, including ones to find the position of
	    the most or least significant set bit, bitfield insertion and
	    extraction, and bit reversal;
	
	  * a general data conversion instruction (CVT) supporting conversion
	    between any two data types supported by this extension; and
	
	  * new instructions to compute the composite of a set of boolean
	    conditions a group of shader threads.
	
	This extension also provides some new capabilities for individual program
	types, including:
	
	  * support for instanced geometry programs, where a geometry program may
	    be run multiple times for each primitive;
	
	  * support for emitting vertices in a geometry program where each vertex
	    emitted may be directed at a specified vertex stream and captured
	    using the ARB_transform_feedback3 extension;
	
	  * support for interpolating an attribute at a programmable offset
	    relative to the pixel center (IPAO), at a programmable sample number
	    (IPAS), or at the fragment's centroid location (IPAC) in a fragment
	    program;
	
	  * support for reading a mask of covered samples in a fragment program;
	
	  * support for reading a point sprite coordinate directly in a fragment
	    program, without overriding a texture coordinate;
	
	  * support for reading patch primitives and per-patch attributes
	    (introduced by ARB_tessellation_shader) in a geometry program; and
	
	  * support for multiple output vectors for a single color output in a
	    fragment program (as used by ARB_blend_func_extended).
	
	This extension also provides optional support for 64-bit-per-component
	variables and 64-bit floating-point arithmetic.  These features are
	supported if and only if "NV_gpu_program_fp64" is found in the extension
	string.
	
	This extension incorporates the memory access operations from the
	NV_shader_buffer_load and NV_parameter_buffer_object2 extensions,
	originally built as add-ons to NV_gpu_program4.  It also provides the
	following new capabilities:
	
	  * support for the features without requiring a separate OPTION keyword;
	
	  * support for indexing into an array of constant buffers using the LDC
	    opcode added by NV_parameter_buffer_object2;
	
	  * support for storing into buffer objects at a specified GPU address
	    using the STORE opcode, an allowing applications to create READ_WRITE
	    and WRITE_ONLY mappings when making a buffer object resident using the
	    API mechanisms in the NV_shader_buffer_store extension;
	
	  * storage instruction modifiers to allow loading and storing 64-bit
	    component values;
	
	  * support for atomic memory transactions using the ATOM opcode, where
	    the instruction atomically reads the memory pointed to by a pointer,
	    performs a specified computation, stores the results of that
	    computation, and returns the original value read;
	
	  * support for memory barrier transactions using the MEMBAR opcode, which
	    ensures that all memory stores issued prior to the opcode complete
	    prior to any subsequent memory transactions; and
	
	  * a fragment program option to specify that depth and stencil tests are
	    performed prior to fragment program execution.
	
	Additionally, the assembly program languages supported by this extension
	include support for reading, writing, and performing atomic memory
	operations on texture image data using the opcodes and mechanisms
	documented in the "Dependencies on NV_gpu_program5" section of the
	EXT_shader_image_load_store extension.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/NV/gpu_program5.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.NV.gpu_program5 import *
from OpenGL.raw.GL.NV.gpu_program5 import _EXTENSION_NAME

def glInitGpuProgram5NV():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

# INPUT glProgramSubroutineParametersuivNV.params size not checked against count
glProgramSubroutineParametersuivNV=wrapper.wrapper(glProgramSubroutineParametersuivNV).setInputArraySize(
    'params', None
)
glGetProgramSubroutineParameteruivNV=wrapper.wrapper(glGetProgramSubroutineParameteruivNV).setOutput(
    'param',size=_glgets._glget_size_mapping,pnameArg='target',orPassIn=True
)
### END AUTOGENERATED SECTION


"""Storage of per-context values of various types

Because OpenGL needs persistent references to the
objects we're constructing to shadow Python objects,
we have to store references to the objects somewhere

For any given Python GUI library, we can use a weakref
to the library's representation of the GL context to 
call the cleanup function.  That means some per-GUI 
library code in OpenGL (or the library), but it gives 
us very natural operations within OpenGL.

Note: you can entirely disable use of this module by 
setting:

    OpenGL.ERROR_ON_COPY = True 
    OpenGL.STORE_POINTERS = False 
        
before importing OpenGL functionality.
"""
from OpenGL import platform
import weakref
storedPointers = {
    # map from contextID: { constant: value }
}
storedWeakPointers = {
    # map from contextID: WeakValueDictionary({ constant: value })
}
STORAGES = [ storedPointers, storedWeakPointers ]

def getContext( context = None ):
    """Get the context (if passed, just return)
    
    context -- the context ID, if None, the current context
    """
    if context is None:
        context = platform.GetCurrentContext()
        if context == 0:
            from OpenGL import error
            raise error.Error(
                """Attempt to retrieve context when no valid context"""
            )
    return context
def setValue( constant, value, context=None, weak=False ):
    """Set a stored value for the given context
    
    constant -- Normally a GL constant value, but can be any hashable value 
    value -- the value to be stored.  If weak is true must be 
        weak-reference-able.  If None, then the value will be deleted from 
        the storage 
    context -- the context identifier for which we're storing the value
    weak -- if true, value will be stored with a weakref
        Note: you should always pass the same value for "weak" for a given 
        constant, otherwise you will create two storages for the constant.
    """
    if getattr( value, '_no_cache_', False ):
        return 
    context = getContext( context )
    if weak:
        storage = storedWeakPointers
        cls = weakref.WeakValueDictionary
    else:
        storage = storedPointers
        cls = dict
    current = storage.get( context )
    if current is None:
        storage[context] = current = cls()
    previous = current.get( constant )
    if value is None:
        try:
            del current[ constant ]
        except (KeyError,TypeError,ValueError) as err:
            pass 
    else:
        # XXX potential for failure here if a non-weakref-able objects
        # is being stored with weak == True
        current[ constant ] = value 
    return previous
def delValue( constant, context=None ):
    """Delete the specified value for the given context
    
    constant -- Normally a GL constant value, but can be any hashable value 
    context -- the context identifier for which we're storing the value
    """
    context = getContext( context )
    found = False
    for storage in STORAGES:
        contextStorage = storage.get( context  )
        if contextStorage:
            try:
                del contextStorage[ constant ]
                found = True
            except KeyError as err:
                pass
    return found

def getValue( constant, context = None ):
    """Get a stored value for the given constant
    
    constant -- unique ID for the type of data being retrieved
    context -- the context ID, if None, the current context
    """
    context = getContext( context )
    for storage in STORAGES:
        contextStorage = storage.get( context  )
        if contextStorage:
            value =  contextStorage.get( constant )
            if value is not None:
                return value
    return None

def cleanupContext( context=None ):
    """Cleanup all held pointer objects for the given context
    
    Warning: this is dangerous, as if you call it before a context 
    is destroyed you may release memory held by the context and cause
    a protection fault when the GL goes to render the scene!
    
    Normally you will want to get the context ID explicitly and then 
    register cleanupContext as a weakref callback to your GUI library 
    Context object with the (now invalid) context ID as parameter.
    """
    if context is None:
        context = platform.GetCurrentContext()
    for storage in STORAGES:
        try:
            del storedPointers[ context ]
        except KeyError as err:
            return False
        else:
            return True



'''OpenGL extension HP.image_transform

This module customises the behaviour of the 
OpenGL.raw.GL.HP.image_transform to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension provides support for scaling, rotation, and translation
	of two-dimensional pixel rectangles at a fixed location in the pixel
	transfer process.  The 2D image transformation attributes are specified
	as individual values so that that implementations may easily detect
	scaling and rotation values that lend themselves to optimization.  2D
	image transformation occurs immediately after the post-convolution color
	table stage of the pixel pipeline.  This extension also defines a color
	table that is applied immediately after the image transformation operation.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/HP/image_transform.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.HP.image_transform import *
from OpenGL.raw.GL.HP.image_transform import _EXTENSION_NAME

def glInitImageTransformHP():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

# INPUT glImageTransformParameterivHP.params size not checked against 'pname'
glImageTransformParameterivHP=wrapper.wrapper(glImageTransformParameterivHP).setInputArraySize(
    'params', None
)
# INPUT glImageTransformParameterfvHP.params size not checked against 'pname'
glImageTransformParameterfvHP=wrapper.wrapper(glImageTransformParameterfvHP).setInputArraySize(
    'params', None
)
glGetImageTransformParameterivHP=wrapper.wrapper(glGetImageTransformParameterivHP).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
glGetImageTransformParameterfvHP=wrapper.wrapper(glGetImageTransformParameterfvHP).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
### END AUTOGENERATED SECTION


'''OpenGL extension EXT.coordinate_frame

This module customises the behaviour of the 
OpenGL.raw.GL.EXT.coordinate_frame to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension allows specifying a per-vertex tangent and binormal
	vector in addition to the normal vector, defining a coordinate frame.
	The coordinate frame is used in additional extensions which also build
	on fragment lighting to achieve bump mapping.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/EXT/coordinate_frame.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.EXT.coordinate_frame import *
from OpenGL.raw.GL.EXT.coordinate_frame import _EXTENSION_NAME

def glInitCoordinateFrameEXT():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

glTangent3bvEXT=wrapper.wrapper(glTangent3bvEXT).setInputArraySize(
    'v', 3
)
glTangent3dvEXT=wrapper.wrapper(glTangent3dvEXT).setInputArraySize(
    'v', 3
)
glTangent3fvEXT=wrapper.wrapper(glTangent3fvEXT).setInputArraySize(
    'v', 3
)
glTangent3ivEXT=wrapper.wrapper(glTangent3ivEXT).setInputArraySize(
    'v', 3
)
glTangent3svEXT=wrapper.wrapper(glTangent3svEXT).setInputArraySize(
    'v', 3
)
glBinormal3bvEXT=wrapper.wrapper(glBinormal3bvEXT).setInputArraySize(
    'v', 3
)
glBinormal3dvEXT=wrapper.wrapper(glBinormal3dvEXT).setInputArraySize(
    'v', 3
)
glBinormal3fvEXT=wrapper.wrapper(glBinormal3fvEXT).setInputArraySize(
    'v', 3
)
glBinormal3ivEXT=wrapper.wrapper(glBinormal3ivEXT).setInputArraySize(
    'v', 3
)
glBinormal3svEXT=wrapper.wrapper(glBinormal3svEXT).setInputArraySize(
    'v', 3
)
# INPUT glTangentPointerEXT.pointer size not checked against 'type,stride'
glTangentPointerEXT=wrapper.wrapper(glTangentPointerEXT).setInputArraySize(
    'pointer', None
)
# INPUT glBinormalPointerEXT.pointer size not checked against 'type,stride'
glBinormalPointerEXT=wrapper.wrapper(glBinormalPointerEXT).setInputArraySize(
    'pointer', None
)
### END AUTOGENERATED SECTION


"""Image-handling routines

### Unresolved:

    Following methods are not yet resolved due to my not being sure how the
    function should be wrapped:

        glCompressedTexImage3D
        glCompressedTexImage2D
        glCompressedTexImage1D
        glCompressedTexSubImage3D
        glCompressedTexSubImage2D
        glCompressedTexSubImage1D
"""
from OpenGL.raw.GL.VERSION import GL_1_1,GL_1_2, GL_3_0
from OpenGL import images, arrays, wrapper
from OpenGL.arrays import arraydatatype
from OpenGL._bytes import bytes,integer_types
from OpenGL.raw.GL import _types
import ctypes

def asInt( value ):
    if isinstance( value, float ):
        return int(round(value,0))
    return value

## update the image tables with standard image types...
images.COMPONENT_COUNTS.update( {
    GL_1_1.GL_BITMAP : 1, # must be GL_UNSIGNED_BYTE

    GL_1_1.GL_RED : 1,
    GL_1_1.GL_GREEN : 1,
    GL_1_1.GL_BLUE : 1,
    GL_1_1.GL_ALPHA : 1,
    GL_3_0.GL_RED_INTEGER : 1,
    GL_3_0.GL_GREEN_INTEGER : 1,
    GL_3_0.GL_BLUE_INTEGER : 1,
    GL_3_0.GL_ALPHA_INTEGER : 1,
    GL_1_1.GL_LUMINANCE : 1,
    GL_1_1.GL_LUMINANCE_ALPHA : 2,
    GL_1_1.GL_COLOR_INDEX : 1,
    GL_1_1.GL_STENCIL_INDEX : 1,
    GL_1_1.GL_DEPTH_COMPONENT : 1,

    GL_1_1.GL_RGB : 3,
    GL_1_2.GL_BGR : 3,
    GL_3_0.GL_RGB16F : 3,
    GL_3_0.GL_RGB16I : 3,
    GL_3_0.GL_RGB16UI : 3,
    GL_3_0.GL_RGB32F : 3,
    GL_3_0.GL_RGB32I : 3,
    GL_3_0.GL_RGB32UI : 3,
    GL_3_0.GL_RGB8I : 3,
    GL_3_0.GL_RGB8UI : 3,
    GL_3_0.GL_RGB9_E5 : 3,
    GL_3_0.GL_RGB_INTEGER : 3,

    GL_1_1.GL_RGBA : 4,
    GL_1_2.GL_BGRA : 4,
    GL_3_0.GL_RGBA16F : 4,
    GL_3_0.GL_RGBA16I : 4,
    GL_3_0.GL_RGBA16UI : 4,
    GL_3_0.GL_RGBA32F : 4,
    GL_3_0.GL_RGBA32I : 4,
    GL_3_0.GL_RGBA32UI : 4,
    GL_3_0.GL_RGBA8I : 4,
    GL_3_0.GL_RGBA8UI : 4,
    GL_3_0.GL_RGBA_INTEGER : 4,
} )

images.TYPE_TO_ARRAYTYPE.update( {
    GL_1_2.GL_UNSIGNED_BYTE_3_3_2 : GL_1_1.GL_UNSIGNED_BYTE,
    GL_1_2.GL_UNSIGNED_BYTE_2_3_3_REV : GL_1_1.GL_UNSIGNED_BYTE,
    GL_1_2.GL_UNSIGNED_SHORT_4_4_4_4 : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_2.GL_UNSIGNED_SHORT_4_4_4_4_REV : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_2.GL_UNSIGNED_SHORT_5_5_5_1 : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_2.GL_UNSIGNED_SHORT_1_5_5_5_REV : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_2.GL_UNSIGNED_SHORT_5_6_5 : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_2.GL_UNSIGNED_SHORT_5_6_5_REV : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_2.GL_UNSIGNED_INT_8_8_8_8 : GL_1_1.GL_UNSIGNED_INT,
    GL_1_2.GL_UNSIGNED_INT_8_8_8_8_REV : GL_1_1.GL_UNSIGNED_INT,
    GL_1_2.GL_UNSIGNED_INT_10_10_10_2 : GL_1_1.GL_UNSIGNED_INT,
    GL_1_2.GL_UNSIGNED_INT_2_10_10_10_REV : GL_1_1.GL_UNSIGNED_INT,
    GL_1_1.GL_UNSIGNED_BYTE : GL_1_1.GL_UNSIGNED_BYTE,
    GL_1_1.GL_BYTE: GL_1_1.GL_BYTE,
    GL_1_1.GL_UNSIGNED_SHORT : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_1.GL_SHORT :  GL_1_1.GL_SHORT,
    GL_1_1.GL_UNSIGNED_INT : GL_1_1.GL_UNSIGNED_INT,
    GL_1_1.GL_INT : GL_1_1.GL_INT,
    GL_1_1.GL_FLOAT : GL_1_1.GL_FLOAT,
    GL_1_1.GL_DOUBLE : GL_1_1.GL_DOUBLE,
    GL_1_1.GL_BITMAP : GL_1_1.GL_UNSIGNED_BYTE,
} )
images.TIGHT_PACK_FORMATS.update({
    GL_1_2.GL_UNSIGNED_BYTE_3_3_2 : 3,
    GL_1_2.GL_UNSIGNED_BYTE_2_3_3_REV : 3,
    GL_1_2.GL_UNSIGNED_SHORT_4_4_4_4 : 4,
    GL_1_2.GL_UNSIGNED_SHORT_4_4_4_4_REV : 4,
    GL_1_2.GL_UNSIGNED_SHORT_5_5_5_1 : 4,
    GL_1_2.GL_UNSIGNED_SHORT_1_5_5_5_REV : 4,
    GL_1_2.GL_UNSIGNED_SHORT_5_6_5 : 3,
    GL_1_2.GL_UNSIGNED_SHORT_5_6_5_REV : 3,
    GL_1_2.GL_UNSIGNED_INT_8_8_8_8 : 4,
    GL_1_2.GL_UNSIGNED_INT_8_8_8_8_REV : 4,
    GL_1_2.GL_UNSIGNED_INT_10_10_10_2 : 4,
    GL_1_2.GL_UNSIGNED_INT_2_10_10_10_REV : 4,
    GL_1_1.GL_BITMAP: 8, # single bits, 8 of them...
})

images.RANK_PACKINGS.update( {
    4: [
        # Note the sgis parameters are skipped here unless you import 
        # the sgis texture4D extension...
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_ALIGNMENT, 1),
    ],
    3: [
        (GL_1_1.glPixelStorei,GL_1_2.GL_PACK_SKIP_IMAGES, 0),
        (GL_1_1.glPixelStorei,GL_1_2.GL_PACK_IMAGE_HEIGHT, 0),
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_ALIGNMENT, 1),
    ],
    2: [
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_ROW_LENGTH, 0),
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_SKIP_ROWS, 0),
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_ALIGNMENT, 1),
    ],
    1: [
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_SKIP_PIXELS, 0),
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_ALIGNMENT, 1),
    ],
} )


__all__ = (
    'glReadPixels',
    'glReadPixelsb',
    'glReadPixelsd',
    'glReadPixelsf',
    'glReadPixelsi',
    'glReadPixelss',
    'glReadPixelsub',
    'glReadPixelsui',
    'glReadPixelsus',

    'glGetTexImage',

    'glDrawPixels',
    'glDrawPixelsb',
    'glDrawPixelsf',
    'glDrawPixelsi',
    'glDrawPixelss',
    'glDrawPixelsub',
    'glDrawPixelsui',
    'glDrawPixelsus',


    'glTexSubImage2D',
    'glTexSubImage1D',
    #'glTexSubImage3D',

    'glTexImage1D',
    'glTexImage2D',
    #'glTexImage3D',

    'glGetTexImageb',
    'glGetTexImaged',
    'glGetTexImagef',
    'glGetTexImagei',
    'glGetTexImages',
    'glGetTexImageub',
    'glGetTexImageui',
    'glGetTexImageus',
    'glTexImage1Db',
    'glTexImage2Db',
    #'glTexImage3Db',
    'glTexSubImage1Db',
    'glTexSubImage2Db',
    #'glTexSubImage3Db',
    'glTexImage1Df',
    'glTexImage2Df',
    #'glTexImage3Df',
    'glTexSubImage1Df',
    'glTexSubImage2Df',
    #'glTexSubImage3Df',
    'glTexImage1Di',
    'glTexImage2Di',
    #'glTexImage3Di',
    'glTexSubImage1Di',
    'glTexSubImage2Di',
    #'glTexSubImage3Di',
    'glTexImage1Ds',
    'glTexImage2Ds',
    #'glTexImage3Ds',
    'glTexSubImage1Ds',
    'glTexSubImage2Ds',
    #'glTexSubImage3Ds',
    'glTexImage1Dub',
    'glTexImage2Dub',
    #'glTexImage3Dub',
    'glTexSubImage1Dub',
    'glTexSubImage2Dub',
    #'glTexSubImage3Dub',
    'glTexImage1Dui',
    'glTexImage2Dui',
    #'glTexImage3Dui',
    'glTexSubImage1Dui',
    'glTexSubImage2Dui',
    #'glTexSubImage3Dui',
    'glTexImage1Dus',
    'glTexImage2Dus',
    #'glTexImage3Dus',
    'glTexSubImage1Dus',
    'glTexSubImage2Dus',
    #'glTexSubImage3Dus',

    #'glColorTable',
    #'glGetColorTable',
    #'glColorSubTable',

    #'glConvolutionFilter1D',
    #'glConvolutionFilter2D',
    #'glGetConvolutionFilter',
    #'glSeparableFilter2D',
    #'glGetSeparableFilter',

    #'glGetMinmax',
)

def _get_texture_level_dims(target,level):
    """Retrieve texture dims for given level and target"""
    dims = []
    dim = _types.GLuint()
    GL_1_1.glGetTexLevelParameteriv( target, level, GL_1_1.GL_TEXTURE_WIDTH, dim )
    dims = [dim.value]
    if target != GL_1_1.GL_TEXTURE_1D:
        GL_1_1.glGetTexLevelParameteriv( target, level, GL_1_1.GL_TEXTURE_HEIGHT, dim )
        dims.append( dim.value )
        if target != GL_1_1.GL_TEXTURE_2D:
            GL_1_1.glGetTexLevelParameteriv( target, level, GL_1_1.GL_TEXTURE_DEPTH, dim )
            dims.append( dim.value )
    return dims

for suffix,type in [
    ('b',GL_1_1.GL_BYTE),
    ('d',GL_1_1.GL_DOUBLE),
    ('f',GL_1_1.GL_FLOAT),
    ('i',GL_1_1.GL_INT),
    ('s',GL_1_1.GL_SHORT),
    ('ub',GL_1_1.GL_UNSIGNED_BYTE),
    ('ui',GL_1_1.GL_UNSIGNED_INT),
    ('us',GL_1_1.GL_UNSIGNED_SHORT),
]:
    def glReadPixels( x,y,width,height,format,type=type, array=None, outputType=bytes ):
        """Read specified pixels from the current display buffer

        This typed version returns data in your specified default
        array data-type format, or in the passed array, which will
        be converted to the array-type required by the format.
        """
        x,y,width,height = asInt(x),asInt(y),asInt(width),asInt(height)
        arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ images.TYPE_TO_ARRAYTYPE.get(type,type) ]
        
        if array is None:
            array = imageData = images.SetupPixelRead( format, (width,height), type )
            owned = True
        else:
            if isinstance( array, integer_types):
                imageData = ctypes.c_void_p( array )
            else:
                array = arrayType.asArray( array )
                imageData = arrayType.voidDataPointer( array )
            owned = False
        GL_1_1.glReadPixels(
            x,y,
            width, height,
            format,type,
            imageData
        )
        if owned and outputType is bytes:
            return images.returnFormat( array, type )
        else:
            return array
    globals()["glReadPixels%s"%(suffix,)] = glReadPixels
    def glGetTexImage( target, level,format,type=type, array=None, outputType=bytes ):
        """Get a texture-level as an image
        
        target -- enum constant for the texture engine to be read
        level -- the mip-map level to read
        format -- image format to read out the data
        type -- data-type into which to read the data
        array -- optional array/offset into which to store the value

        outputType -- default (bytes) provides string output of the
            results iff OpenGL.UNSIGNED_BYTE_IMAGES_AS_STRING is True
            and type == GL_UNSIGNED_BYTE.  Any other value will cause
            output in the default array output format.

        returns the pixel data array in the format defined by the
        format, type and outputType
        """
        arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ images.TYPE_TO_ARRAYTYPE.get(type,type) ]
        if array is None:
            dims = _get_texture_level_dims(target,level)
            array = imageData = images.SetupPixelRead( format, tuple(dims), type )
            owned = True
        else:
            if isinstance( array, integer_types):
                imageData = ctypes.c_void_p( array )
            else:
                array = arrayType.asArray( array )
                imageData = arrayType.voidDataPointer( array )
            owned = False
        GL_1_1.glGetTexImage(
            target, level, format, type, imageData
        )
        if owned and outputType is bytes:
            return images.returnFormat( array, type )
        else:
            return array
    globals()["glGetTexImage%s"%(suffix,)] = glGetTexImage
##	def glGetTexSubImage( target, level,format,type ):
##		"""Get a texture-level as an image"""
##		dims = [GL_1_1.glGetTexLevelParameteriv( target, level, GL_1_1.GL_TEXTURE_WIDTH )]
##		if target != GL_1_1.GL_TEXTURE_1D:
##			dims.append( GL_1_1.glGetTexLevelParameteriv( target, level, GL_1_1.GL_TEXTURE_HEIGHT ) )
##			if target != GL_1_1.GL_TEXTURE_2D:
##				dims.append( GL_1_1.glGetTexLevelParameteriv( target, level, GL_1_2.GL_TEXTURE_DEPTH ) )
##		array = images.SetupPixelRead( format, tuple(dims), type )
##		arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ images.TYPE_TO_ARRAYTYPE.get(type,type) ]
##		GL_1_1.glGetTexImage(
##			target, level, format, type, ctypes.c_void_p( arrayType.dataPointer(array))
##		)
##		return array
##	"%s = glGetTexImage"%(suffix)
    try:
        del suffix,type
    except NameError as err:
        pass
# Now the real glReadPixels...
def glReadPixels( x,y,width,height,format,type, array=None, outputType=bytes ):
    """Read specified pixels from the current display buffer

    x,y,width,height -- location and dimensions of the image to read
        from the buffer
    format -- pixel format for the resulting data
    type -- data-format for the resulting data
    array -- optional array/offset into which to store the value
    outputType -- default (bytes) provides string output of the
        results iff OpenGL.UNSIGNED_BYTE_IMAGES_AS_STRING is True
        and type == GL_UNSIGNED_BYTE.  Any other value will cause
        output in the default array output format.

    returns the pixel data array in the format defined by the
    format, type and outputType
    """
    x,y,width,height = asInt(x),asInt(y),asInt(width),asInt(height)

    arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ images.TYPE_TO_ARRAYTYPE.get(type,type) ]
    if array is None:
        array = imageData = images.SetupPixelRead( format, (width,height), type )
        owned = True
    else:
        if isinstance( array, integer_types):
            imageData = ctypes.c_void_p( array )
        else:
            array = arrayType.asArray( array )
            imageData = arrayType.voidDataPointer( array )
        owned = False

    GL_1_1.glReadPixels(
        x,y,width,height,
        format,type,
        imageData
    )
    if owned and outputType is bytes:
        return images.returnFormat( array, type )
    else:
        return array

def glGetTexImage( target, level,format,type, array=None, outputType=bytes ):
    """Get a texture-level as an image

    target -- enum constant for the texture engine to be read
    level -- the mip-map level to read
    format -- image format to read out the data
    type -- data-type into which to read the data
    array -- optional array/offset into which to store the value

    outputType -- default (bytes) provides string output of the
        results iff OpenGL.UNSIGNED_BYTE_IMAGES_AS_STRING is True
        and type == GL_UNSIGNED_BYTE.  Any other value will cause
        output in the default array output format.

    returns the pixel data array in the format defined by the
    format, type and outputType
    """
    arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ images.TYPE_TO_ARRAYTYPE.get(type,type) ]
    if array is None:
        dims = _get_texture_level_dims(target,level)
        array = imageData = images.SetupPixelRead( format, tuple(dims), type )
        owned = True
    else:
        if isinstance( array, integer_types):
            imageData = ctypes.c_void_p( array )
        else:
            array = arrayType.asArray( array )
            imageData = arrayType.voidDataPointer( array )
        owned = False
    GL_1_1.glGetTexImage(
        target, level, format, type, imageData
    )
    if outputType is bytes:
        return images.returnFormat( array, type )
    else:
        return array


INT_DIMENSION_NAMES = [
    'width','height','depth','x','y','z',
    'xoffset','yoffset','zoffset',
    'start', 'count',
]
def asWrapper( value ):
    if not isinstance( value, wrapper.Wrapper ):
        return wrapper.wrapper( value )
    return value

def asIntConverter( value, *args ):
    if isinstance( value, float ):
        return int(round(value,0))
    return value

def setDimensionsAsInts( baseOperation ):
    """Set arguments with names in INT_DIMENSION_NAMES to asInt processing"""
    baseOperation = asWrapper( baseOperation )
    argNames = getattr( baseOperation, 'pyConverterNames', baseOperation.argNames )
    for i,argName in enumerate(argNames):
        if argName in INT_DIMENSION_NAMES:
            baseOperation.setPyConverter( argName, asIntConverter )
    return baseOperation



class ImageInputConverter( object ):
    def __init__( self, rank, pixelsName=None, typeName='type' ):
        self.rank = rank
        self.typeName = typeName
        self.pixelsName = pixelsName
    def finalise( self, wrapper ):
        """Get our pixel index from the wrapper"""
        self.typeIndex = wrapper.pyArgIndex( self.typeName )
        self.pixelsIndex = wrapper.pyArgIndex( self.pixelsName )
    def __call__( self, arg, baseOperation, pyArgs ):
        """pyConverter for the pixels argument"""
        images.setupDefaultTransferMode()
        images.rankPacking( self.rank )
        type = pyArgs[ self.typeIndex ]
        arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ images.TYPE_TO_ARRAYTYPE[ type ] ]
        return arrayType.asArray( arg )
#	def cResolver( self, array ):
#		return array
#		return ctypes.c_void_p( arrays.ArrayDatatype.dataPointer( array ) )

class TypedImageInputConverter( ImageInputConverter ):
    def __init__( self, rank, pixelsName, arrayType, typeName=None ):
        self.rank = rank
        self.arrayType = arrayType
        self.pixelsName = pixelsName
        self.typeName = typeName
    def __call__( self, arg, baseOperation, pyArgs ):
        """The pyConverter for the pixels"""
        images.setupDefaultTransferMode()
        images.rankPacking( self.rank )
        return self.arrayType.asArray( arg )
    def finalise( self, wrapper ):
        """Get our pixel index from the wrapper"""
        self.pixelsIndex = wrapper.pyArgIndex( self.pixelsName )
    def width( self, pyArgs, index, wrappedOperation ):
        """Extract the width from the pixels argument"""
        return self.arrayType.dimensions( pyArgs[self.pixelsIndex] )[0]
    def height( self, pyArgs, index, wrappedOperation ):
        """Extract the height from the pixels argument"""
        return self.arrayType.dimensions( pyArgs[self.pixelsIndex] )[1]
    def depth( self, pyArgs, index, wrappedOperation ):
        """Extract the depth from the pixels argument"""
        return self.arrayType.dimensions( pyArgs[self.pixelsIndex] )[2]
    def type( self, pyArgs, index, wrappedOperation ):
        """Provide the item-type argument from our stored value

        This is used for pre-bound processing where we want to provide
        the type by implication...
        """
        return self.typeName

class CompressedImageConverter( object ):
    def finalise( self, wrapper ):
        """Get our pixel index from the wrapper"""
        self.dataIndex = wrapper.pyArgIndex( 'data' )
    def __call__( self, pyArgs, index, wrappedOperation ):
        """Create a data-size measurement for our image"""
        arg = pyArgs[ self.dataIndex ]
        return arraydatatype.ArrayDatatype.arrayByteCount(arg)



DIMENSION_NAMES = (
    'width','height','depth'
)
PIXEL_NAMES = (
    'pixels', 'row', 'column',
)
DATA_SIZE_NAMES = (
    'imageSize',
)

def setImageInput(
    baseOperation, arrayType=None, dimNames=DIMENSION_NAMES,
    pixelName="pixels", typeName=None
):
    """Determine how to convert "pixels" into an image-compatible argument"""
    baseOperation = asWrapper( baseOperation )
    # rank is the count of width,height,depth arguments...
    rank = len([
        # rank is the number of dims we want, not the number we give...
        argName for argName in baseOperation.argNames
        if argName in dimNames
    ]) + 1
    if arrayType:
        converter = TypedImageInputConverter( rank, pixelName, arrayType, typeName=typeName )
        for i,argName in enumerate(baseOperation.argNames):
            if argName in dimNames:
                baseOperation.setPyConverter( argName )
                baseOperation.setCConverter( argName, getattr(converter,argName) )
            elif argName == 'type' and typeName is not None:
                baseOperation.setPyConverter( argName )
                baseOperation.setCConverter( argName, converter.type )
    else:
        converter = ImageInputConverter( rank, pixelsName=pixelName, typeName=typeName or 'type' )
    for argName in baseOperation.argNames:
        if argName in DATA_SIZE_NAMES:
            baseOperation.setPyConverter( argName )
            baseOperation.setCConverter( argName, converter.imageDataSize )
    baseOperation.setPyConverter(
        pixelName, converter,
    )
#	baseOperation.setCResolver(
#		pixelName, converter.cResolver
#	)
    return baseOperation

glDrawPixels = setDimensionsAsInts(
    setImageInput(
        GL_1_1.glDrawPixels
    )
)
glTexSubImage2D = setDimensionsAsInts(
    setImageInput(
        GL_1_1.glTexSubImage2D
    )
)
glTexSubImage1D = setDimensionsAsInts(
    setImageInput(
        GL_1_1.glTexSubImage1D
    )
)
glTexImage2D = setDimensionsAsInts(
    setImageInput(
        GL_1_1.glTexImage2D
    )
)
glTexImage1D = setDimensionsAsInts(
    setImageInput(
        GL_1_1.glTexImage1D
    )
)

def typedImageFunction( suffix, arrayConstant,  baseFunction ):
    """Produce a typed version of the given image function"""
    functionName = baseFunction.__name__
    functionName = '%(functionName)s%(suffix)s'%locals()
    if baseFunction:
        arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ arrayConstant ]
        function = setDimensionsAsInts(
            setImageInput(
                baseFunction,
                arrayType,
                typeName = arrayConstant,
            )
        )
        return functionName, function
    else:
        return functionName, baseFunction

def _setDataSize( baseFunction, argument='imageSize' ):
    """Set the data-size value to come from the data field"""
    if baseFunction:
        converter = CompressedImageConverter()
        return asWrapper( baseFunction ).setPyConverter(
            argument
        ).setCConverter( argument, converter )
    else:
        return baseFunction

def compressedImageFunction( baseFunction ):
    """Set the imageSize and dimensions-as-ints converters for baseFunction"""
    if baseFunction:
        return setDimensionsAsInts(
            _setDataSize(
                baseFunction, argument='imageSize'
            )
        )
    else:
        return baseFunction

for suffix,arrayConstant in [
    ('b', GL_1_1.GL_BYTE),
    ('f', GL_1_1.GL_FLOAT),
    ('i', GL_1_1.GL_INT),
    ('s', GL_1_1.GL_SHORT),
    ('ub', GL_1_1.GL_UNSIGNED_BYTE),
    ('ui', GL_1_1.GL_UNSIGNED_INT),
    ('us', GL_1_1.GL_UNSIGNED_SHORT),
]:
    for functionName in (
        'glTexImage1D','glTexImage2D',
        'glTexSubImage1D','glTexSubImage2D',
        'glDrawPixels',
        #'glTexSubImage3D','glTexImage3D', # extension/1.2 standard
    ):
        functionName, function = typedImageFunction(
            suffix, arrayConstant, getattr(GL_1_1,functionName),
        )
        globals()[functionName] = function
        try:
            del function, functionName
        except NameError as err:
            pass
    try:
        del suffix,arrayConstant
    except NameError as err:
        pass



"""OSX specific extensions to GLUT"""
from OpenGL import platform
from OpenGL import constant
from OpenGL.GLUT import special

GLUT_NO_RECOVERY = constant.Constant( 'GLUT_NO_RECOVERY', 1024)
GLUT_3_2_CORE_PROFILE = constant.Constant( 'GLUT_3_2_CORE_PROFILE', 2048)

glutCheckLoop = platform.createBaseFunction( 
    'glutCheckLoop', dll=platform.PLATFORM.GLUT, resultType=None, 
    argTypes=[],
    doc='glutCheckLoop(  ) -> None', 
    argNames=(),
)

glutWMCloseFunc = special.GLUTCallback(
    'WMClose', (), (),
)



'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GL import _types as _cs
# End users want this...
from OpenGL.raw.GL._types import *
from OpenGL.raw.GL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GL_ATI_texture_float'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GL,'GL_ATI_texture_float',error_checker=_errors._error_checker)
GL_ALPHA_FLOAT16_ATI=_C('GL_ALPHA_FLOAT16_ATI',0x881C)
GL_ALPHA_FLOAT32_ATI=_C('GL_ALPHA_FLOAT32_ATI',0x8816)
GL_INTENSITY_FLOAT16_ATI=_C('GL_INTENSITY_FLOAT16_ATI',0x881D)
GL_INTENSITY_FLOAT32_ATI=_C('GL_INTENSITY_FLOAT32_ATI',0x8817)
GL_LUMINANCE_ALPHA_FLOAT16_ATI=_C('GL_LUMINANCE_ALPHA_FLOAT16_ATI',0x881F)
GL_LUMINANCE_ALPHA_FLOAT32_ATI=_C('GL_LUMINANCE_ALPHA_FLOAT32_ATI',0x8819)
GL_LUMINANCE_FLOAT16_ATI=_C('GL_LUMINANCE_FLOAT16_ATI',0x881E)
GL_LUMINANCE_FLOAT32_ATI=_C('GL_LUMINANCE_FLOAT32_ATI',0x8818)
GL_RGBA_FLOAT16_ATI=_C('GL_RGBA_FLOAT16_ATI',0x881A)
GL_RGBA_FLOAT32_ATI=_C('GL_RGBA_FLOAT32_ATI',0x8814)
GL_RGB_FLOAT16_ATI=_C('GL_RGB_FLOAT16_ATI',0x881B)
GL_RGB_FLOAT32_ATI=_C('GL_RGB_FLOAT32_ATI',0x8815)




'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GL import _types as _cs
# End users want this...
from OpenGL.raw.GL._types import *
from OpenGL.raw.GL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GL_EXT_draw_buffers2'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GL,'GL_EXT_draw_buffers2',error_checker=_errors._error_checker)

@_f
@_p.types(None,_cs.GLuint,_cs.GLboolean,_cs.GLboolean,_cs.GLboolean,_cs.GLboolean)
def glColorMaskIndexedEXT(index,r,g,b,a):pass
@_f
@_p.types(None,_cs.GLenum,_cs.GLuint)
def glDisableIndexedEXT(target,index):pass
@_f
@_p.types(None,_cs.GLenum,_cs.GLuint)
def glEnableIndexedEXT(target,index):pass
@_f
@_p.types(None,_cs.GLenum,_cs.GLuint,arrays.GLbooleanArray)
def glGetBooleanIndexedvEXT(target,index,data):pass
@_f
@_p.types(None,_cs.GLenum,_cs.GLuint,arrays.GLintArray)
def glGetIntegerIndexedvEXT(target,index,data):pass
@_f
@_p.types(_cs.GLboolean,_cs.GLenum,_cs.GLuint)
def glIsEnabledIndexedEXT(target,index):pass



'''OpenGL extension DFX.tbuffer

This module customises the behaviour of the 
OpenGL.raw.GL.DFX.tbuffer to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension allows a write mask to be defined for the fragment
	mask which is created during multisample rendering. This can be used
	to create effects such as motion blur and depth of field.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/DFX/tbuffer.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.DFX.tbuffer import *
from OpenGL.raw.GL.DFX.tbuffer import _EXTENSION_NAME

def glInitTbufferDFX():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )


### END AUTOGENERATED SECTION


'''OpenGL extension NV.geometry_shader4

This module customises the behaviour of the 
OpenGL.raw.GL.NV.geometry_shader4 to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension builds upon the EXT_geometry_shader4 specification to
	provide two additional capabilities:
	
	    * Support for QUADS, QUAD_STRIP, and POLYGON primitive types when
	      geometry shaders are enabled.  Such primitives will be tessellated
	      into individual triangles.
	
	    * Setting the value of GEOMETRY_VERTICES_OUT_EXT will take effect
	      immediately.  It is not necessary to link the program object in
	      order for this change to take effect, as is the case in the EXT
	      version of this extension.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/NV/geometry_shader4.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.NV.geometry_shader4 import *
from OpenGL.raw.GL.NV.geometry_shader4 import _EXTENSION_NAME

def glInitGeometryShader4NV():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )


### END AUTOGENERATED SECTION


'''OpenGL extension SUNX.constant_data

This module customises the behaviour of the 
OpenGL.raw.GL.SUNX.constant_data to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension allows the pixel data specified by the
	application to be used internally without making a second copy.
	This extension affects how the pixel data in client memory is
	interpreted and therefore affects DrawPixels, Bitmap,
	PolygonStipple, TexImage1D, TexImage2D, TexImage3DEXT,
	ColorTableSGI.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/SUNX/constant_data.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.SUNX.constant_data import *
from OpenGL.raw.GL.SUNX.constant_data import _EXTENSION_NAME

def glInitConstantDataSUNX():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )


### END AUTOGENERATED SECTION


'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.EGL import _types as _cs
# End users want this...
from OpenGL.raw.EGL._types import *
from OpenGL.raw.EGL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'EGL_KHR_config_attribs'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.EGL,'EGL_KHR_config_attribs',error_checker=_errors._error_checker)
EGL_CONFORMANT_KHR=_C('EGL_CONFORMANT_KHR',0x3042)
EGL_VG_ALPHA_FORMAT_PRE_BIT_KHR=_C('EGL_VG_ALPHA_FORMAT_PRE_BIT_KHR',0x0040)
EGL_VG_COLORSPACE_LINEAR_BIT_KHR=_C('EGL_VG_COLORSPACE_LINEAR_BIT_KHR',0x0020)




'''OpenGL extension EXT.index_func

This module customises the behaviour of the 
OpenGL.raw.GL.EXT.index_func to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension provides a way to discard fragments when a comparison
	between the fragment's index value and a reference index fails.  This
	may be used similarly to the alpha test which is available in RGBA mode.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/EXT/index_func.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.EXT.index_func import *
from OpenGL.raw.GL.EXT.index_func import _EXTENSION_NAME

def glInitIndexFuncEXT():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )


### END AUTOGENERATED SECTION


'''OpenGL extension NV.register_combiners2

This module customises the behaviour of the 
OpenGL.raw.GL.NV.register_combiners2 to provide a more 
Python-friendly API

Overview (from the spec)
	
	The NV_register_combiners extension provides a powerful fragment
	coloring mechanism.  This specification extends the register combiners
	functionality to support more color constant values that are unique
	for each general combiner stage.
	
	The base register combiners functionality supports only two color
	constants.  These two constants are available in every general
	combiner stage and in the final combiner.
	
	When many general combiner stages are supported, more than two
	unique color constants is often required.  The obvious way to extend
	the register combiners is to add several more color constant
	registers.  But adding new unique color constant registers is
	expensive for hardware implementation because every color constant
	register must be available as an input to any stage.
	
	In practice however, it is the total set of general combiner stages
	that requires more color constants, not each and every individual
	general combiner stage.  Each individual general combiner stage
	typically requires only one or two color constants.
	
	By keeping two color constant registers but making these two registers
	contain two unique color constant values for each general combiner
	stage, the hardware expense of supporting multiple color constants
	is minimized.  Additionally, this scheme scales appropriately as
	more general combiner stages are added.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/NV/register_combiners2.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.NV.register_combiners2 import *
from OpenGL.raw.GL.NV.register_combiners2 import _EXTENSION_NAME

def glInitRegisterCombiners2NV():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

# INPUT glCombinerStageParameterfvNV.params size not checked against 'pname'
glCombinerStageParameterfvNV=wrapper.wrapper(glCombinerStageParameterfvNV).setInputArraySize(
    'params', None
)
glGetCombinerStageParameterfvNV=wrapper.wrapper(glGetCombinerStageParameterfvNV).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
### END AUTOGENERATED SECTION


'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GL import _types as _cs
# End users want this...
from OpenGL.raw.GL._types import *
from OpenGL.raw.GL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GL_NV_texture_multisample'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GL,'GL_NV_texture_multisample',error_checker=_errors._error_checker)
GL_TEXTURE_COLOR_SAMPLES_NV=_C('GL_TEXTURE_COLOR_SAMPLES_NV',0x9046)
GL_TEXTURE_COVERAGE_SAMPLES_NV=_C('GL_TEXTURE_COVERAGE_SAMPLES_NV',0x9045)
@_f
@_p.types(None,_cs.GLenum,_cs.GLsizei,_cs.GLsizei,_cs.GLint,_cs.GLsizei,_cs.GLsizei,_cs.GLboolean)
def glTexImage2DMultisampleCoverageNV(target,coverageSamples,colorSamples,internalFormat,width,height,fixedSampleLocations):pass
@_f
@_p.types(None,_cs.GLenum,_cs.GLsizei,_cs.GLsizei,_cs.GLint,_cs.GLsizei,_cs.GLsizei,_cs.GLsizei,_cs.GLboolean)
def glTexImage3DMultisampleCoverageNV(target,coverageSamples,colorSamples,internalFormat,width,height,depth,fixedSampleLocations):pass
@_f
@_p.types(None,_cs.GLuint,_cs.GLenum,_cs.GLsizei,_cs.GLsizei,_cs.GLint,_cs.GLsizei,_cs.GLsizei,_cs.GLboolean)
def glTextureImage2DMultisampleCoverageNV(texture,target,coverageSamples,colorSamples,internalFormat,width,height,fixedSampleLocations):pass
@_f
@_p.types(None,_cs.GLuint,_cs.GLenum,_cs.GLsizei,_cs.GLint,_cs.GLsizei,_cs.GLsizei,_cs.GLboolean)
def glTextureImage2DMultisampleNV(texture,target,samples,internalFormat,width,height,fixedSampleLocations):pass
@_f
@_p.types(None,_cs.GLuint,_cs.GLenum,_cs.GLsizei,_cs.GLsizei,_cs.GLint,_cs.GLsizei,_cs.GLsizei,_cs.GLsizei,_cs.GLboolean)
def glTextureImage3DMultisampleCoverageNV(texture,target,coverageSamples,colorSamples,internalFormat,width,height,depth,fixedSampleLocations):pass
@_f
@_p.types(None,_cs.GLuint,_cs.GLenum,_cs.GLsizei,_cs.GLint,_cs.GLsizei,_cs.GLsizei,_cs.GLsizei,_cs.GLboolean)
def glTextureImage3DMultisampleNV(texture,target,samples,internalFormat,width,height,depth,fixedSampleLocations):pass



'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.EGL import _types as _cs
# End users want this...
from OpenGL.raw.EGL._types import *
from OpenGL.raw.EGL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'EGL_KHR_lock_surface'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.EGL,'EGL_KHR_lock_surface',error_checker=_errors._error_checker)
EGL_BITMAP_ORIGIN_KHR=_C('EGL_BITMAP_ORIGIN_KHR',0x30C8)
EGL_BITMAP_PITCH_KHR=_C('EGL_BITMAP_PITCH_KHR',0x30C7)
EGL_BITMAP_PIXEL_ALPHA_OFFSET_KHR=_C('EGL_BITMAP_PIXEL_ALPHA_OFFSET_KHR',0x30CC)
EGL_BITMAP_PIXEL_BLUE_OFFSET_KHR=_C('EGL_BITMAP_PIXEL_BLUE_OFFSET_KHR',0x30CB)
EGL_BITMAP_PIXEL_GREEN_OFFSET_KHR=_C('EGL_BITMAP_PIXEL_GREEN_OFFSET_KHR',0x30CA)
EGL_BITMAP_PIXEL_LUMINANCE_OFFSET_KHR=_C('EGL_BITMAP_PIXEL_LUMINANCE_OFFSET_KHR',0x30CD)
EGL_BITMAP_PIXEL_RED_OFFSET_KHR=_C('EGL_BITMAP_PIXEL_RED_OFFSET_KHR',0x30C9)
EGL_BITMAP_POINTER_KHR=_C('EGL_BITMAP_POINTER_KHR',0x30C6)
EGL_FORMAT_RGBA_8888_EXACT_KHR=_C('EGL_FORMAT_RGBA_8888_EXACT_KHR',0x30C2)
EGL_FORMAT_RGBA_8888_KHR=_C('EGL_FORMAT_RGBA_8888_KHR',0x30C3)
EGL_FORMAT_RGB_565_EXACT_KHR=_C('EGL_FORMAT_RGB_565_EXACT_KHR',0x30C0)
EGL_FORMAT_RGB_565_KHR=_C('EGL_FORMAT_RGB_565_KHR',0x30C1)
EGL_LOCK_SURFACE_BIT_KHR=_C('EGL_LOCK_SURFACE_BIT_KHR',0x0080)
EGL_LOCK_USAGE_HINT_KHR=_C('EGL_LOCK_USAGE_HINT_KHR',0x30C5)
EGL_LOWER_LEFT_KHR=_C('EGL_LOWER_LEFT_KHR',0x30CE)
EGL_MAP_PRESERVE_PIXELS_KHR=_C('EGL_MAP_PRESERVE_PIXELS_KHR',0x30C4)
EGL_MATCH_FORMAT_KHR=_C('EGL_MATCH_FORMAT_KHR',0x3043)
EGL_OPTIMAL_FORMAT_BIT_KHR=_C('EGL_OPTIMAL_FORMAT_BIT_KHR',0x0100)
EGL_READ_SURFACE_BIT_KHR=_C('EGL_READ_SURFACE_BIT_KHR',0x0001)
EGL_UPPER_LEFT_KHR=_C('EGL_UPPER_LEFT_KHR',0x30CF)
EGL_WRITE_SURFACE_BIT_KHR=_C('EGL_WRITE_SURFACE_BIT_KHR',0x0002)
@_f
@_p.types(_cs.EGLBoolean,_cs.EGLDisplay,_cs.EGLSurface,arrays.GLintArray)
def eglLockSurfaceKHR(dpy,surface,attrib_list):pass
@_f
@_p.types(_cs.EGLBoolean,_cs.EGLDisplay,_cs.EGLSurface)
def eglUnlockSurfaceKHR(dpy,surface):pass



'''OpenGL extension OES.fixed_point

This module customises the behaviour of the 
OpenGL.raw.GLES1.OES.fixed_point to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/OES/fixed_point.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GLES1 import _types, _glgets
from OpenGL.raw.GLES1.OES.fixed_point import *
from OpenGL.raw.GLES1.OES.fixed_point import _EXTENSION_NAME

def glInitFixedPointOES():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

glClipPlanexOES=wrapper.wrapper(glClipPlanexOES).setInputArraySize(
    'equation', 4
)
# INPUT glFogxvOES.param size not checked against 'pname'
glFogxvOES=wrapper.wrapper(glFogxvOES).setInputArraySize(
    'param', None
)
glGetClipPlanexOES=wrapper.wrapper(glGetClipPlanexOES).setOutput(
    'equation',size=(4,),orPassIn=True
)
glGetFixedvOES=wrapper.wrapper(glGetFixedvOES).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
glGetTexEnvxvOES=wrapper.wrapper(glGetTexEnvxvOES).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
glGetTexParameterxvOES=wrapper.wrapper(glGetTexParameterxvOES).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
# INPUT glLightModelxvOES.param size not checked against 'pname'
glLightModelxvOES=wrapper.wrapper(glLightModelxvOES).setInputArraySize(
    'param', None
)
# INPUT glLightxvOES.params size not checked against 'pname'
glLightxvOES=wrapper.wrapper(glLightxvOES).setInputArraySize(
    'params', None
)
glLoadMatrixxOES=wrapper.wrapper(glLoadMatrixxOES).setInputArraySize(
    'm', 16
)
# INPUT glMaterialxvOES.param size not checked against 'pname'
glMaterialxvOES=wrapper.wrapper(glMaterialxvOES).setInputArraySize(
    'param', None
)
glMultMatrixxOES=wrapper.wrapper(glMultMatrixxOES).setInputArraySize(
    'm', 16
)
# INPUT glPointParameterxvOES.params size not checked against 'pname'
glPointParameterxvOES=wrapper.wrapper(glPointParameterxvOES).setInputArraySize(
    'params', None
)
# INPUT glTexEnvxvOES.params size not checked against 'pname'
glTexEnvxvOES=wrapper.wrapper(glTexEnvxvOES).setInputArraySize(
    'params', None
)
# INPUT glTexParameterxvOES.params size not checked against 'pname'
glTexParameterxvOES=wrapper.wrapper(glTexParameterxvOES).setInputArraySize(
    'params', None
)
# INPUT glGetLightxvOES.params size not checked against 'pname'
glGetLightxvOES=wrapper.wrapper(glGetLightxvOES).setInputArraySize(
    'params', None
)
# INPUT glGetMaterialxvOES.params size not checked against 'pname'
glGetMaterialxvOES=wrapper.wrapper(glGetMaterialxvOES).setInputArraySize(
    'params', None
)
# INPUT glBitmapxOES.bitmap size not checked against 'width,height'
glBitmapxOES=wrapper.wrapper(glBitmapxOES).setInputArraySize(
    'bitmap', None
)
glColor3xvOES=wrapper.wrapper(glColor3xvOES).setInputArraySize(
    'components', 3
)
glColor4xvOES=wrapper.wrapper(glColor4xvOES).setInputArraySize(
    'components', 4
)
# INPUT glConvolutionParameterxvOES.params size not checked against 'pname'
glConvolutionParameterxvOES=wrapper.wrapper(glConvolutionParameterxvOES).setInputArraySize(
    'params', None
)
glEvalCoord1xvOES=wrapper.wrapper(glEvalCoord1xvOES).setInputArraySize(
    'coords', 1
)
glEvalCoord2xvOES=wrapper.wrapper(glEvalCoord2xvOES).setInputArraySize(
    'coords', 2
)
# INPUT glFeedbackBufferxOES.buffer size not checked against n
glFeedbackBufferxOES=wrapper.wrapper(glFeedbackBufferxOES).setInputArraySize(
    'buffer', None
)
glGetConvolutionParameterxvOES=wrapper.wrapper(glGetConvolutionParameterxvOES).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
glGetHistogramParameterxvOES=wrapper.wrapper(glGetHistogramParameterxvOES).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
glGetLightxOES=wrapper.wrapper(glGetLightxOES).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
glGetMapxvOES=wrapper.wrapper(glGetMapxvOES).setOutput(
    'v',size=_glgets._glget_size_mapping,pnameArg='query',orPassIn=True
)
glGetPixelMapxv=wrapper.wrapper(glGetPixelMapxv).setOutput(
    'values',size=lambda x:(x,),pnameArg='size',orPassIn=True
)
glGetTexGenxvOES=wrapper.wrapper(glGetTexGenxvOES).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
glGetTexLevelParameterxvOES=wrapper.wrapper(glGetTexLevelParameterxvOES).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
glIndexxvOES=wrapper.wrapper(glIndexxvOES).setInputArraySize(
    'component', 1
)
glLoadTransposeMatrixxOES=wrapper.wrapper(glLoadTransposeMatrixxOES).setInputArraySize(
    'm', 16
)
glMultTransposeMatrixxOES=wrapper.wrapper(glMultTransposeMatrixxOES).setInputArraySize(
    'm', 16
)
glMultiTexCoord1xvOES=wrapper.wrapper(glMultiTexCoord1xvOES).setInputArraySize(
    'coords', 1
)
glMultiTexCoord2xvOES=wrapper.wrapper(glMultiTexCoord2xvOES).setInputArraySize(
    'coords', 2
)
glMultiTexCoord3xvOES=wrapper.wrapper(glMultiTexCoord3xvOES).setInputArraySize(
    'coords', 3
)
glMultiTexCoord4xvOES=wrapper.wrapper(glMultiTexCoord4xvOES).setInputArraySize(
    'coords', 4
)
glNormal3xvOES=wrapper.wrapper(glNormal3xvOES).setInputArraySize(
    'coords', 3
)
# INPUT glPixelMapx.values size not checked against size
glPixelMapx=wrapper.wrapper(glPixelMapx).setInputArraySize(
    'values', None
)
# INPUT glPrioritizeTexturesxOES.priorities size not checked against n
# INPUT glPrioritizeTexturesxOES.textures size not checked against n
glPrioritizeTexturesxOES=wrapper.wrapper(glPrioritizeTexturesxOES).setInputArraySize(
    'priorities', None
).setInputArraySize(
    'textures', None
)
glRasterPos2xvOES=wrapper.wrapper(glRasterPos2xvOES).setInputArraySize(
    'coords', 2
)
glRasterPos3xvOES=wrapper.wrapper(glRasterPos3xvOES).setInputArraySize(
    'coords', 3
)
glRasterPos4xvOES=wrapper.wrapper(glRasterPos4xvOES).setInputArraySize(
    'coords', 4
)
glRectxvOES=wrapper.wrapper(glRectxvOES).setInputArraySize(
    'v1', 2
).setInputArraySize(
    'v2', 2
)
glTexCoord1xvOES=wrapper.wrapper(glTexCoord1xvOES).setInputArraySize(
    'coords', 1
)
glTexCoord2xvOES=wrapper.wrapper(glTexCoord2xvOES).setInputArraySize(
    'coords', 2
)
glTexCoord3xvOES=wrapper.wrapper(glTexCoord3xvOES).setInputArraySize(
    'coords', 3
)
glTexCoord4xvOES=wrapper.wrapper(glTexCoord4xvOES).setInputArraySize(
    'coords', 4
)
# INPUT glTexGenxvOES.params size not checked against 'pname'
glTexGenxvOES=wrapper.wrapper(glTexGenxvOES).setInputArraySize(
    'params', None
)
glVertex2xvOES=wrapper.wrapper(glVertex2xvOES).setInputArraySize(
    'coords', 2
)
glVertex3xvOES=wrapper.wrapper(glVertex3xvOES).setInputArraySize(
    'coords', 3
)
glVertex4xvOES=wrapper.wrapper(glVertex4xvOES).setInputArraySize(
    'coords', 4
)
### END AUTOGENERATED SECTION


'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GL import _types as _cs
# End users want this...
from OpenGL.raw.GL._types import *
from OpenGL.raw.GL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GL_ARB_vertex_blend'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GL,'GL_ARB_vertex_blend',error_checker=_errors._error_checker)
GL_ACTIVE_VERTEX_UNITS_ARB=_C('GL_ACTIVE_VERTEX_UNITS_ARB',0x86A5)
GL_CURRENT_WEIGHT_ARB=_C('GL_CURRENT_WEIGHT_ARB',0x86A8)
GL_MAX_VERTEX_UNITS_ARB=_C('GL_MAX_VERTEX_UNITS_ARB',0x86A4)
GL_MODELVIEW0_ARB=_C('GL_MODELVIEW0_ARB',0x1700)
GL_MODELVIEW10_ARB=_C('GL_MODELVIEW10_ARB',0x872A)
GL_MODELVIEW11_ARB=_C('GL_MODELVIEW11_ARB',0x872B)
GL_MODELVIEW12_ARB=_C('GL_MODELVIEW12_ARB',0x872C)
GL_MODELVIEW13_ARB=_C('GL_MODELVIEW13_ARB',0x872D)
GL_MODELVIEW14_ARB=_C('GL_MODELVIEW14_ARB',0x872E)
GL_MODELVIEW15_ARB=_C('GL_MODELVIEW15_ARB',0x872F)
GL_MODELVIEW16_ARB=_C('GL_MODELVIEW16_ARB',0x8730)
GL_MODELVIEW17_ARB=_C('GL_MODELVIEW17_ARB',0x8731)
GL_MODELVIEW18_ARB=_C('GL_MODELVIEW18_ARB',0x8732)
GL_MODELVIEW19_ARB=_C('GL_MODELVIEW19_ARB',0x8733)
GL_MODELVIEW1_ARB=_C('GL_MODELVIEW1_ARB',0x850A)
GL_MODELVIEW20_ARB=_C('GL_MODELVIEW20_ARB',0x8734)
GL_MODELVIEW21_ARB=_C('GL_MODELVIEW21_ARB',0x8735)
GL_MODELVIEW22_ARB=_C('GL_MODELVIEW22_ARB',0x8736)
GL_MODELVIEW23_ARB=_C('GL_MODELVIEW23_ARB',0x8737)
GL_MODELVIEW24_ARB=_C('GL_MODELVIEW24_ARB',0x8738)
GL_MODELVIEW25_ARB=_C('GL_MODELVIEW25_ARB',0x8739)
GL_MODELVIEW26_ARB=_C('GL_MODELVIEW26_ARB',0x873A)
GL_MODELVIEW27_ARB=_C('GL_MODELVIEW27_ARB',0x873B)
GL_MODELVIEW28_ARB=_C('GL_MODELVIEW28_ARB',0x873C)
GL_MODELVIEW29_ARB=_C('GL_MODELVIEW29_ARB',0x873D)
GL_MODELVIEW2_ARB=_C('GL_MODELVIEW2_ARB',0x8722)
GL_MODELVIEW30_ARB=_C('GL_MODELVIEW30_ARB',0x873E)
GL_MODELVIEW31_ARB=_C('GL_MODELVIEW31_ARB',0x873F)
GL_MODELVIEW3_ARB=_C('GL_MODELVIEW3_ARB',0x8723)
GL_MODELVIEW4_ARB=_C('GL_MODELVIEW4_ARB',0x8724)
GL_MODELVIEW5_ARB=_C('GL_MODELVIEW5_ARB',0x8725)
GL_MODELVIEW6_ARB=_C('GL_MODELVIEW6_ARB',0x8726)
GL_MODELVIEW7_ARB=_C('GL_MODELVIEW7_ARB',0x8727)
GL_MODELVIEW8_ARB=_C('GL_MODELVIEW8_ARB',0x8728)
GL_MODELVIEW9_ARB=_C('GL_MODELVIEW9_ARB',0x8729)
GL_VERTEX_BLEND_ARB=_C('GL_VERTEX_BLEND_ARB',0x86A7)
GL_WEIGHT_ARRAY_ARB=_C('GL_WEIGHT_ARRAY_ARB',0x86AD)
GL_WEIGHT_ARRAY_POINTER_ARB=_C('GL_WEIGHT_ARRAY_POINTER_ARB',0x86AC)
GL_WEIGHT_ARRAY_SIZE_ARB=_C('GL_WEIGHT_ARRAY_SIZE_ARB',0x86AB)
GL_WEIGHT_ARRAY_STRIDE_ARB=_C('GL_WEIGHT_ARRAY_STRIDE_ARB',0x86AA)
GL_WEIGHT_ARRAY_TYPE_ARB=_C('GL_WEIGHT_ARRAY_TYPE_ARB',0x86A9)
GL_WEIGHT_SUM_UNITY_ARB=_C('GL_WEIGHT_SUM_UNITY_ARB',0x86A6)
@_f
@_p.types(None,_cs.GLint)
def glVertexBlendARB(count):pass
@_f
@_p.types(None,_cs.GLint,_cs.GLenum,_cs.GLsizei,ctypes.c_void_p)
def glWeightPointerARB(size,type,stride,pointer):pass
@_f
@_p.types(None,_cs.GLint,arrays.GLbyteArray)
def glWeightbvARB(size,weights):pass
@_f
@_p.types(None,_cs.GLint,arrays.GLdoubleArray)
def glWeightdvARB(size,weights):pass
@_f
@_p.types(None,_cs.GLint,arrays.GLfloatArray)
def glWeightfvARB(size,weights):pass
@_f
@_p.types(None,_cs.GLint,arrays.GLintArray)
def glWeightivARB(size,weights):pass
@_f
@_p.types(None,_cs.GLint,arrays.GLshortArray)
def glWeightsvARB(size,weights):pass
@_f
@_p.types(None,_cs.GLint,arrays.GLubyteArray)
def glWeightubvARB(size,weights):pass
@_f
@_p.types(None,_cs.GLint,arrays.GLuintArray)
def glWeightuivARB(size,weights):pass
@_f
@_p.types(None,_cs.GLint,arrays.GLushortArray)
def glWeightusvARB(size,weights):pass



"""glGet* auto-generation of output arrays (DO NOT EDIT, AUTOGENERATED)"""
try:
    from OpenGL.raw.GL._lookupint import LookupInt as _L
except ImportError:
    def _L(*args):
        raise RuntimeError( "Need to define a lookupint for this api" )
_glget_size_mapping = _m = {}




'''OpenGL extension ARB.shader_draw_parameters

This module customises the behaviour of the 
OpenGL.raw.GL.ARB.shader_draw_parameters to provide a more 
Python-friendly API

Overview (from the spec)
	
	    In unextended GL, vertex shaders have inputs named gl_VertexID and
	gl_InstanceID, which contain, respectively the index of the vertex and
	instance. The value of gl_VertexID is the implicitly passed index of the
	vertex being processed, which includes the value of baseVertex, for those
	commands that accept it. Meanwhile, gl_InstanceID is the integer index
	of the current instance being processed, but, even for commands that
	accept a baseInstance parameter, it does not include the value of this
	argument. Furthermore, the equivalents to these variables in other
	graphics APIs do not necessarily follow these conventions. The reason for
	this inconsistency is that there are legitimate use cases for both
	inclusion and exclusion of the baseVertex or baseInstance parameters
	in gl_VertexID and gl_InstanceID, respectively.
	
	    Rather than change the semantics of either built-in variable, this
	extension adds two new built-in variables to the GL shading language,
	gl_BaseVertexARB and gl_BaseInstanceARB, which contain the values passed
	in the baseVertex and baseInstance parameters, respectively. Shaders
	provided by the application may use these variables to offset gl_VertexID
	or gl_InstanceID if desired, or use them for any other purpose.
	
	    Additionally, this extension adds a further built-in variable, gl_DrawID
	to the shading language. This variable contains the index of the draw
	currently being processed by a Multi* variant of a drawing command (such
	as MultiDrawElements or MultiDrawArraysIndirect).

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/ARB/shader_draw_parameters.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.ARB.shader_draw_parameters import *
from OpenGL.raw.GL.ARB.shader_draw_parameters import _EXTENSION_NAME

def glInitShaderDrawParametersARB():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )


### END AUTOGENERATED SECTION


'''OpenGL extension SGIX.ycrcb

This module customises the behaviour of the 
OpenGL.raw.GL.SGIX.ycrcb to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension provides a method for OpenGL to read and store 
	images that are defined in standard YCRCB 422 and 444 video formats.
	As with the CYMK extension, conversion to RGBA takes place immediately 
	following the unpack pixel store, and preceding the pack pixel store 
	operations, and is only available on transfers to and from memory.  
	The algorithms that convert between YCRCB and RGBA are "black-box"
	in nature, and left undefined by the extension.
	
	Two new formats are added, YCRCB_422_SGIX and YCRCB_444_SGIX.
	
	To handle the difference in sampling rate for 422 video, the pixel 
	storage operations treat YCRCB_422_SGIX as a 2 component format,
	where the first component represents chroma, and the second luma.
	The chroma component alternates between Cb and Cr values on
	a per pixel basis.  If the specified image <width> parameter is not
	a multiple of 2, then fragments or texels that result from processing 
	the <width>th column of pixels will have undefined color value.
	
	YCRCB_444_SGIX is defined as a 3 component format representing 
	the Cb, Y, and Cr values per pixel.
	
	As with the CMYK extension, this extension doesn't preclude the 
	possiblity of other higher quality conversion methods.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/SGIX/ycrcb.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.SGIX.ycrcb import *
from OpenGL.raw.GL.SGIX.ycrcb import _EXTENSION_NAME

def glInitYcrcbSGIX():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )


### END AUTOGENERATED SECTION
from OpenGL import images as _i
_i.COMPONENT_COUNTS[ GL_YCRCB_422_SGIX ] = 1 # must be GL_UNSIGNED_BYTE
_i.COMPONENT_COUNTS[ GL_YCRCB_444_SGIX ] = 1 # must be GL_UNSIGNED_SHORT



'''OpenGL extension ARB.ES2_compatibility

This module customises the behaviour of the 
OpenGL.raw.GL.ARB.ES2_compatibility to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension adds support for features of OpenGL ES 2.0 that are
	missing from OpenGL 3.x. Enabling these features will ease the process
	of porting applications from OpenGL ES 2.0 to OpenGL.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/ARB/ES2_compatibility.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.ARB.ES2_compatibility import *
from OpenGL.raw.GL.ARB.ES2_compatibility import _EXTENSION_NAME

def glInitEs2CompatibilityARB():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

# INPUT glShaderBinary.binary size not checked against length
# INPUT glShaderBinary.shaders size not checked against count
glShaderBinary=wrapper.wrapper(glShaderBinary).setInputArraySize(
    'binary', None
).setInputArraySize(
    'shaders', None
)
glGetShaderPrecisionFormat=wrapper.wrapper(glGetShaderPrecisionFormat).setOutput(
    'precision',size=(2,),orPassIn=True
).setOutput(
    'range',size=(2,),orPassIn=True
)
### END AUTOGENERATED SECTION
from OpenGL import lazywrapper as _lazywrapper
from OpenGL.arrays import GLintArray
@_lazywrapper.lazy( glGetShaderPrecisionFormat )
def glGetShaderPrecisionFormat(baseOperation, shadertype, precisiontype, range=None,precision=None ):
    """Provides range and precision if not provided, returns (range,precision)"""
    if range is None:
        range = GLintArray.zeros( (2,))
    if precision is None:
        precision = GLintArray.zeros((2,))
    baseOperation( shadertype, precisiontype, range, precision )
    return range, precision



"""Common code for accelerated modules"""
import logging
from OpenGL import _configflags
needed_version = (3,1,0)
_log = logging.getLogger( 'OpenGL.acceleratesupport' )
try:
    import OpenGL_accelerate
    if _configflags.USE_ACCELERATE:
        if OpenGL_accelerate.__version_tuple__ <  needed_version:
            _log.warning( """Incompatible version of OpenGL_accelerate found, need at least %s found %s""", needed_version, OpenGL_accelerate.__version_tuple__)
            raise ImportError( """Old version of OpenGL_accelerate""" )
        ACCELERATE_AVAILABLE = True
        _log.info( """OpenGL_accelerate module loaded""" )
    else:
        raise ImportError( """Acceleration disabled""" )
except ImportError as err:
    _log.info( """No OpenGL_accelerate module loaded: %s""", err )
    ACCELERATE_AVAILABLE = False



'''OpenGL extension VERSION.GL_3_0

This module customises the behaviour of the 
OpenGL.raw.GL.VERSION.GL_3_0 to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/VERSION/GL_3_0.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.VERSION.GL_3_0 import *
from OpenGL.raw.GL.VERSION.GL_3_0 import _EXTENSION_NAME

def glInitGl30VERSION():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

glGetBooleani_v=wrapper.wrapper(glGetBooleani_v).setOutput(
    'data',size=_glgets._glget_size_mapping,pnameArg='target',orPassIn=True
)
glGetIntegeri_v=wrapper.wrapper(glGetIntegeri_v).setOutput(
    'data',size=_glgets._glget_size_mapping,pnameArg='target',orPassIn=True
)
# INPUT glTransformFeedbackVaryings.varyings size not checked against count
glTransformFeedbackVaryings=wrapper.wrapper(glTransformFeedbackVaryings).setInputArraySize(
    'varyings', None
)
glGetTransformFeedbackVarying=wrapper.wrapper(glGetTransformFeedbackVarying).setOutput(
    'length',size=(1,),orPassIn=True
).setOutput(
    'name',size=lambda x:(x,),pnameArg='bufSize',orPassIn=True
).setOutput(
    'size',size=(1,),orPassIn=True
).setOutput(
    'type',size=(1,),orPassIn=True
)
# INPUT glVertexAttribIPointer.pointer size not checked against 'size,type,stride'
glVertexAttribIPointer=wrapper.wrapper(glVertexAttribIPointer).setInputArraySize(
    'pointer', None
)
glGetVertexAttribIiv=wrapper.wrapper(glGetVertexAttribIiv).setOutput(
    'params',size=(1,),orPassIn=True
)
glGetVertexAttribIuiv=wrapper.wrapper(glGetVertexAttribIuiv).setOutput(
    'params',size=(1,),orPassIn=True
)
glVertexAttribI1iv=wrapper.wrapper(glVertexAttribI1iv).setInputArraySize(
    'v', 1
)
glVertexAttribI2iv=wrapper.wrapper(glVertexAttribI2iv).setInputArraySize(
    'v', 2
)
glVertexAttribI3iv=wrapper.wrapper(glVertexAttribI3iv).setInputArraySize(
    'v', 3
)
glVertexAttribI4iv=wrapper.wrapper(glVertexAttribI4iv).setInputArraySize(
    'v', 4
)
glVertexAttribI1uiv=wrapper.wrapper(glVertexAttribI1uiv).setInputArraySize(
    'v', 1
)
glVertexAttribI2uiv=wrapper.wrapper(glVertexAttribI2uiv).setInputArraySize(
    'v', 2
)
glVertexAttribI3uiv=wrapper.wrapper(glVertexAttribI3uiv).setInputArraySize(
    'v', 3
)
glVertexAttribI4uiv=wrapper.wrapper(glVertexAttribI4uiv).setInputArraySize(
    'v', 4
)
glVertexAttribI4bv=wrapper.wrapper(glVertexAttribI4bv).setInputArraySize(
    'v', 4
)
glVertexAttribI4sv=wrapper.wrapper(glVertexAttribI4sv).setInputArraySize(
    'v', 4
)
glVertexAttribI4ubv=wrapper.wrapper(glVertexAttribI4ubv).setInputArraySize(
    'v', 4
)
glVertexAttribI4usv=wrapper.wrapper(glVertexAttribI4usv).setInputArraySize(
    'v', 4
)
# OUTPUT glGetUniformuiv.params COMPSIZE(program, location) 
# INPUT glBindFragDataLocation.name size not checked against 'name'
glBindFragDataLocation=wrapper.wrapper(glBindFragDataLocation).setInputArraySize(
    'name', None
)
# INPUT glGetFragDataLocation.name size not checked against 'name'
glGetFragDataLocation=wrapper.wrapper(glGetFragDataLocation).setInputArraySize(
    'name', None
)
# INPUT glUniform1uiv.value size not checked against count
glUniform1uiv=wrapper.wrapper(glUniform1uiv).setInputArraySize(
    'value', None
)
# INPUT glUniform2uiv.value size not checked against count*2
glUniform2uiv=wrapper.wrapper(glUniform2uiv).setInputArraySize(
    'value', None
)
# INPUT glUniform3uiv.value size not checked against count*3
glUniform3uiv=wrapper.wrapper(glUniform3uiv).setInputArraySize(
    'value', None
)
# INPUT glUniform4uiv.value size not checked against count*4
glUniform4uiv=wrapper.wrapper(glUniform4uiv).setInputArraySize(
    'value', None
)
# INPUT glTexParameterIiv.params size not checked against 'pname'
glTexParameterIiv=wrapper.wrapper(glTexParameterIiv).setInputArraySize(
    'params', None
)
# INPUT glTexParameterIuiv.params size not checked against 'pname'
glTexParameterIuiv=wrapper.wrapper(glTexParameterIuiv).setInputArraySize(
    'params', None
)
glGetTexParameterIiv=wrapper.wrapper(glGetTexParameterIiv).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
glGetTexParameterIuiv=wrapper.wrapper(glGetTexParameterIuiv).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
# INPUT glClearBufferiv.value size not checked against 'buffer'
glClearBufferiv=wrapper.wrapper(glClearBufferiv).setInputArraySize(
    'value', None
)
# INPUT glClearBufferuiv.value size not checked against 'buffer'
glClearBufferuiv=wrapper.wrapper(glClearBufferuiv).setInputArraySize(
    'value', None
)
# INPUT glClearBufferfv.value size not checked against 'buffer'
glClearBufferfv=wrapper.wrapper(glClearBufferfv).setInputArraySize(
    'value', None
)
# INPUT glDeleteRenderbuffers.renderbuffers size not checked against n
glDeleteRenderbuffers=wrapper.wrapper(glDeleteRenderbuffers).setInputArraySize(
    'renderbuffers', None
)
glGenRenderbuffers=wrapper.wrapper(glGenRenderbuffers).setOutput(
    'renderbuffers',size=lambda x:(x,),pnameArg='n',orPassIn=True
)
glGetRenderbufferParameteriv=wrapper.wrapper(glGetRenderbufferParameteriv).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
# INPUT glDeleteFramebuffers.framebuffers size not checked against n
glDeleteFramebuffers=wrapper.wrapper(glDeleteFramebuffers).setInputArraySize(
    'framebuffers', None
)
glGenFramebuffers=wrapper.wrapper(glGenFramebuffers).setOutput(
    'framebuffers',size=lambda x:(x,),pnameArg='n',orPassIn=True
)
glGetFramebufferAttachmentParameteriv=wrapper.wrapper(glGetFramebufferAttachmentParameteriv).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
# INPUT glDeleteVertexArrays.arrays size not checked against n
glDeleteVertexArrays=wrapper.wrapper(glDeleteVertexArrays).setInputArraySize(
    'arrays', None
)
glGenVertexArrays=wrapper.wrapper(glGenVertexArrays).setOutput(
    'arrays',size=lambda x:(x,),pnameArg='n',orPassIn=True
)
### END AUTOGENERATED SECTION
from ctypes import c_char_p
glGetStringi.restype = c_char_p



'''OpenGL extension NV.transform_feedback

This module customises the behaviour of the 
OpenGL.raw.GL.NV.transform_feedback to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension provides a new mode to the GL, called transform feedback,
	which records vertex attributes of the primitives processed by the GL.
	The selected attributes are written into buffer objects, and can be
	written with each attribute in a separate buffer object or with all
	attributes interleaved into a single buffer object.  If a geometry program
	or shader is active, the primitives recorded are those emitted by the
	geometry program.  Otherwise, transform feedback captures primitives whose
	vertex are transformed by a vertex program or shader, or by fixed-function
	vertex processing.  In either case, the primitives captured are those
	generated prior to clipping.  Transform feedback mode is capable of
	capturing transformed vertex data generated by fixed-function vertex
	processing, outputs from assembly vertex or geometry programs, or varying
	variables emitted from GLSL vertex or geometry shaders.
	
	The vertex data recorded in transform feedback mode is stored into buffer
	objects as an array of vertex attributes.  The regular representation and
	the use of buffer objects allows the recorded data to be processed
	directly by the GL without requiring CPU intervention to copy data.  In
	particular, transform feedback data can be used for vertex arrays (via
	vertex buffer objects), as the source for pixel data (via pixel buffer
	objects), as program constant data (via the NV_parameter_buffer_object or
	EXT_bindable_uniform extension), or via any other extension that makes use
	of buffer objects.
	
	This extension introduces new query object support to allow transform
	feedback mode to operate asynchronously.  Query objects allow applications
	to determine when transform feedback results are complete, as well as the
	number of primitives processed and written back to buffer objects while in
	transform feedback mode.  This extension also provides a new rasterizer
	discard enable, which allows applications to use transform feedback to
	capture vertex attributes without rendering anything.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/NV/transform_feedback.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.NV.transform_feedback import *
from OpenGL.raw.GL.NV.transform_feedback import _EXTENSION_NAME

def glInitTransformFeedbackNV():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

# INPUT glTransformFeedbackAttribsNV.attribs size not checked against 'count'
glTransformFeedbackAttribsNV=wrapper.wrapper(glTransformFeedbackAttribsNV).setInputArraySize(
    'attribs', None
)
# INPUT glTransformFeedbackVaryingsNV.locations size not checked against count
glTransformFeedbackVaryingsNV=wrapper.wrapper(glTransformFeedbackVaryingsNV).setInputArraySize(
    'locations', None
)
# INPUT glActiveVaryingNV.name size not checked against 'name'
glActiveVaryingNV=wrapper.wrapper(glActiveVaryingNV).setInputArraySize(
    'name', None
)
# INPUT glGetVaryingLocationNV.name size not checked against 'name'
glGetVaryingLocationNV=wrapper.wrapper(glGetVaryingLocationNV).setInputArraySize(
    'name', None
)
# OUTPUT glGetActiveVaryingNV.name COMPSIZE(program, index, bufSize) 
glGetActiveVaryingNV=wrapper.wrapper(glGetActiveVaryingNV).setOutput(
    'length',size=(1,),orPassIn=True
).setOutput(
    'size',size=(1,),orPassIn=True
).setOutput(
    'type',size=(1,),orPassIn=True
)
glGetTransformFeedbackVaryingNV=wrapper.wrapper(glGetTransformFeedbackVaryingNV).setOutput(
    'location',size=(1,),orPassIn=True
)
# INPUT glTransformFeedbackStreamAttribsNV.attribs size not checked against count
# INPUT glTransformFeedbackStreamAttribsNV.bufstreams size not checked against nbuffers
glTransformFeedbackStreamAttribsNV=wrapper.wrapper(glTransformFeedbackStreamAttribsNV).setInputArraySize(
    'attribs', None
).setInputArraySize(
    'bufstreams', None
)
### END AUTOGENERATED SECTION


'''OpenGL extension EXT.blend_equation_separate

This module customises the behaviour of the 
OpenGL.raw.GL.EXT.blend_equation_separate to provide a more 
Python-friendly API

Overview (from the spec)
	
	EXT_blend_func_separate introduced separate RGB and alpha blend
	factors.  EXT_blend_minmax introduced a distinct blend equation for
	combining source and destination blend terms.  (EXT_blend_subtract &
	EXT_blend_logic_op added other blend equation modes.)  OpenGL 1.4
	integrated both functionalities into the core standard.
	
	While there are separate blend functions for the RGB and alpha blend
	factors, OpenGL 1.4 provides a single blend equation that applies
	to both RGB and alpha portions of blending.
	
	This extension provides a separate blend equation for RGB and alpha
	to match the generality available for blend factors.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/EXT/blend_equation_separate.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.EXT.blend_equation_separate import *
from OpenGL.raw.GL.EXT.blend_equation_separate import _EXTENSION_NAME

def glInitBlendEquationSeparateEXT():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )


### END AUTOGENERATED SECTION


'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GLES2 import _types as _cs
# End users want this...
from OpenGL.raw.GLES2._types import *
from OpenGL.raw.GLES2 import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GLES2_NV_fbo_color_attachments'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GLES2,'GLES2_NV_fbo_color_attachments',error_checker=_errors._error_checker)
GL_COLOR_ATTACHMENT0_NV=_C('GL_COLOR_ATTACHMENT0_NV',0x8CE0)
GL_COLOR_ATTACHMENT10_NV=_C('GL_COLOR_ATTACHMENT10_NV',0x8CEA)
GL_COLOR_ATTACHMENT11_NV=_C('GL_COLOR_ATTACHMENT11_NV',0x8CEB)
GL_COLOR_ATTACHMENT12_NV=_C('GL_COLOR_ATTACHMENT12_NV',0x8CEC)
GL_COLOR_ATTACHMENT13_NV=_C('GL_COLOR_ATTACHMENT13_NV',0x8CED)
GL_COLOR_ATTACHMENT14_NV=_C('GL_COLOR_ATTACHMENT14_NV',0x8CEE)
GL_COLOR_ATTACHMENT15_NV=_C('GL_COLOR_ATTACHMENT15_NV',0x8CEF)
GL_COLOR_ATTACHMENT1_NV=_C('GL_COLOR_ATTACHMENT1_NV',0x8CE1)
GL_COLOR_ATTACHMENT2_NV=_C('GL_COLOR_ATTACHMENT2_NV',0x8CE2)
GL_COLOR_ATTACHMENT3_NV=_C('GL_COLOR_ATTACHMENT3_NV',0x8CE3)
GL_COLOR_ATTACHMENT4_NV=_C('GL_COLOR_ATTACHMENT4_NV',0x8CE4)
GL_COLOR_ATTACHMENT5_NV=_C('GL_COLOR_ATTACHMENT5_NV',0x8CE5)
GL_COLOR_ATTACHMENT6_NV=_C('GL_COLOR_ATTACHMENT6_NV',0x8CE6)
GL_COLOR_ATTACHMENT7_NV=_C('GL_COLOR_ATTACHMENT7_NV',0x8CE7)
GL_COLOR_ATTACHMENT8_NV=_C('GL_COLOR_ATTACHMENT8_NV',0x8CE8)
GL_COLOR_ATTACHMENT9_NV=_C('GL_COLOR_ATTACHMENT9_NV',0x8CE9)
GL_MAX_COLOR_ATTACHMENTS_NV=_C('GL_MAX_COLOR_ATTACHMENTS_NV',0x8CDF)




"""ctypes single-value "parameter" arguments (e.g. byref)
"""
REGISTRY_NAME = 'ctypeparameter'
import ctypes, _ctypes
from OpenGL.raw.GL import _types 
from OpenGL.arrays import _arrayconstants as GL_1_1
from OpenGL import constant
from OpenGL.arrays import formathandler
from OpenGL._bytes import bytes,unicode,as_8_bit
import operator

c = ctypes.c_float(0)
ParamaterType = ctypes.byref(c).__class__
DIRECT_RETURN_TYPES = (
    ParamaterType,
    # these pointer types are implemented as _SimpleCData
    # despite being pointers...
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_wchar_p,
)
try:
    del c
except NameError as err:
    pass

class CtypesParameterHandler( formathandler.FormatHandler ):
    """Ctypes Paramater-type-specific data-type handler for OpenGL"""
    isOutput = True
    HANDLED_TYPES = (ParamaterType, _ctypes._SimpleCData)
    def from_param( cls, value, typeCode=None ):
        if isinstance( value, DIRECT_RETURN_TYPES ):
            return value
        else:
            return ctypes.byref( value )
    from_param = voidDataPointer = classmethod( from_param )
    def dataPointer( cls, value ):
        if isinstance( value, DIRECT_RETURN_TYPES ):
            return value
        else:
            return ctypes.addressof( value )
    dataPointer = classmethod( dataPointer )
    def zeros( self, dims, typeCode ):
        """Return Numpy array of zeros in given size"""
        type = GL_TYPE_TO_ARRAY_MAPPING[ typeCode ]
        for dim in dims:
            type *= dim
        return type() # should expicitly set to 0s
    def ones( self, dims, typeCode='d' ):
        """Return numpy array of ones in given size"""
        raise NotImplementedError( """Haven't got a good ones implementation yet""" )
##		type = GL_TYPE_TO_ARRAY_MAPPING[ typeCode ]
##		for dim in dims:
##			type *= dim
##		return type() # should expicitly set to 0s
    def arrayToGLType( self, value ):
        """Given a value, guess OpenGL type of the corresponding pointer"""
        if isinstance( value, ParamaterType ):
            value = value._obj
        result = ARRAY_TO_GL_TYPE_MAPPING.get( value._type_ )
        if result is not None:
            return result
        raise TypeError(
            """Don't know GL type for array of type %r, known types: %s\nvalue:%s"""%(
                value._type_, list(ARRAY_TO_GL_TYPE_MAPPING.keys()), value,
            )
        )
    def arraySize( self, value, typeCode = None ):
        """Given a data-value, calculate dimensions for the array"""
        if isinstance( value, ParamaterType ):
            value = value._obj
        dims = 1
        for base in self.types( value ):
            length = getattr( base, '_length_', None)
            if length is not None:
                dims *= length
        return dims
    def arrayByteCount( self, value, typeCode = None ):
        """Given a data-value, calculate number of bytes required to represent"""
        if isinstance( value, ParamaterType ):
            value = value._obj
        return ctypes.sizeof( value )
    def types( self, value ):
        """Produce iterable producing all composite types"""
        if isinstance( value, ParamaterType ):
            value = value._obj
        dimObject = value
        while dimObject is not None:
            yield dimObject
            dimObject = getattr( dimObject, '_type_', None )
            if isinstance( dimObject, (bytes,unicode)):
                dimObject = None
    def dims( self, value ):
        """Produce iterable of all dimensions"""
        if isinstance( value, ParamaterType ):
            value = value._obj
        for base in self.types( value ):
            length = getattr( base, '_length_', None)
            if length is not None:
                yield length
    def asArray( self, value, typeCode=None ):
        """Convert given value to an array value of given typeCode"""
        if isinstance( value, DIRECT_RETURN_TYPES ):
            return value
        if isinstance( value, ParamaterType ):
            value = value._obj
        return ctypes.byref( value )
    def unitSize( self, value, typeCode=None ):
        """Determine unit size of an array (if possible)"""
        if isinstance( value, ParamaterType ):
            value = value._obj
        return tuple(self.dims(value))[-1]
    def dimensions( self, value, typeCode=None ):
        """Determine dimensions of the passed array value (if possible)"""
        if isinstance( value, ParamaterType ):
            value = value._obj
        return tuple( self.dims(value) )


ARRAY_TO_GL_TYPE_MAPPING = {
    _types.GLdouble: GL_1_1.GL_DOUBLE,
    _types.GLfloat: GL_1_1.GL_FLOAT,
    _types.GLint: GL_1_1.GL_INT,
    _types.GLuint: GL_1_1.GL_UNSIGNED_INT,
    _types.GLshort: GL_1_1.GL_SHORT,
    _types.GLushort: GL_1_1.GL_UNSIGNED_SHORT,

    _types.GLchar: GL_1_1.GL_CHAR,
    _types.GLbyte: GL_1_1.GL_BYTE,
    _types.GLubyte: GL_1_1.GL_UNSIGNED_BYTE,
}
GL_TYPE_TO_ARRAY_MAPPING = {
    GL_1_1.GL_DOUBLE: _types.GLdouble,
    GL_1_1.GL_FLOAT: _types.GLfloat,
    GL_1_1.GL_INT: _types.GLint,
    GL_1_1.GL_UNSIGNED_INT: _types.GLuint,
    GL_1_1.GL_SHORT: _types.GLshort,
    GL_1_1.GL_UNSIGNED_SHORT: _types.GLushort,

    GL_1_1.GL_CHAR: _types.GLchar,
    GL_1_1.GL_BYTE: _types.GLbyte,
    GL_1_1.GL_UNSIGNED_BYTE: _types.GLubyte,
}



"""Version 1.2 Image-handling functions

Almost all of the 1.2 enhancements are image-handling-related,
so this is, most of the 1.2 wrapper code...

Note that the functions that manually wrap certain operations are
guarded by if simple.functionName checks, so that you can use
if functionName to see if the function is available at run-time.
"""
from OpenGL import wrapper, constants, arrays
from OpenGL.raw.GL.ARB import imaging
from OpenGL.raw.GL.VERSION import GL_1_2 as _simple
from OpenGL.GL.ARB.imaging import *

from OpenGL.GL import images
import ctypes

for suffix,arrayConstant in [
    ('b', constants.GL_BYTE),
    ('f', constants.GL_FLOAT),
    ('i', constants.GL_INT),
    ('s', constants.GL_SHORT),
    ('ub', constants.GL_UNSIGNED_BYTE),
    ('ui', constants.GL_UNSIGNED_INT),
    ('us', constants.GL_UNSIGNED_SHORT),
]:
    for functionName in (
        'glTexImage3D',
        'glTexSubImage3D', # extension/1.2 standard
    ):
        functionName, function = images.typedImageFunction(
            suffix, arrayConstant, getattr(_simple, functionName),
        )
        globals()[functionName] = function
        try:
            del function, functionName
        except NameError as err:
            pass
    try:
        del suffix,arrayConstant
    except NameError as err:
        pass

glTexImage3D = images.setDimensionsAsInts(
    images.setImageInput(
        _simple.glTexImage3D,
        typeName = 'type',
    )
)
glTexSubImage3D = images.setDimensionsAsInts(
    images.setImageInput(
        _simple.glTexSubImage3D,
        typeName = 'type',
    )
)



'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GL import _types as _cs
# End users want this...
from OpenGL.raw.GL._types import *
from OpenGL.raw.GL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GL_SGIX_async'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GL,'GL_SGIX_async',error_checker=_errors._error_checker)
GL_ASYNC_MARKER_SGIX=_C('GL_ASYNC_MARKER_SGIX',0x8329)
@_f
@_p.types(None,_cs.GLuint)
def glAsyncMarkerSGIX(marker):pass
@_f
@_p.types(None,_cs.GLuint,_cs.GLsizei)
def glDeleteAsyncMarkersSGIX(marker,range):pass
@_f
@_p.types(_cs.GLint,arrays.GLuintArray)
def glFinishAsyncSGIX(markerp):pass
@_f
@_p.types(_cs.GLuint,_cs.GLsizei)
def glGenAsyncMarkersSGIX(range):pass
@_f
@_p.types(_cs.GLboolean,_cs.GLuint)
def glIsAsyncMarkerSGIX(marker):pass
@_f
@_p.types(_cs.GLint,arrays.GLuintArray)
def glPollAsyncSGIX(markerp):pass



'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GL import _types as _cs
# End users want this...
from OpenGL.raw.GL._types import *
from OpenGL.raw.GL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GL_PGI_vertex_hints'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GL,'GL_PGI_vertex_hints',error_checker=_errors._error_checker)
GL_COLOR3_BIT_PGI=_C('GL_COLOR3_BIT_PGI',0x00010000)
GL_COLOR4_BIT_PGI=_C('GL_COLOR4_BIT_PGI',0x00020000)
GL_EDGEFLAG_BIT_PGI=_C('GL_EDGEFLAG_BIT_PGI',0x00040000)
GL_INDEX_BIT_PGI=_C('GL_INDEX_BIT_PGI',0x00080000)
GL_MATERIAL_SIDE_HINT_PGI=_C('GL_MATERIAL_SIDE_HINT_PGI',0x1A22C)
GL_MAT_AMBIENT_AND_DIFFUSE_BIT_PGI=_C('GL_MAT_AMBIENT_AND_DIFFUSE_BIT_PGI',0x00200000)
GL_MAT_AMBIENT_BIT_PGI=_C('GL_MAT_AMBIENT_BIT_PGI',0x00100000)
GL_MAT_COLOR_INDEXES_BIT_PGI=_C('GL_MAT_COLOR_INDEXES_BIT_PGI',0x01000000)
GL_MAT_DIFFUSE_BIT_PGI=_C('GL_MAT_DIFFUSE_BIT_PGI',0x00400000)
GL_MAT_EMISSION_BIT_PGI=_C('GL_MAT_EMISSION_BIT_PGI',0x00800000)
GL_MAT_SHININESS_BIT_PGI=_C('GL_MAT_SHININESS_BIT_PGI',0x02000000)
GL_MAT_SPECULAR_BIT_PGI=_C('GL_MAT_SPECULAR_BIT_PGI',0x04000000)
GL_MAX_VERTEX_HINT_PGI=_C('GL_MAX_VERTEX_HINT_PGI',0x1A22D)
GL_NORMAL_BIT_PGI=_C('GL_NORMAL_BIT_PGI',0x08000000)
GL_TEXCOORD1_BIT_PGI=_C('GL_TEXCOORD1_BIT_PGI',0x10000000)
GL_TEXCOORD2_BIT_PGI=_C('GL_TEXCOORD2_BIT_PGI',0x20000000)
GL_TEXCOORD3_BIT_PGI=_C('GL_TEXCOORD3_BIT_PGI',0x40000000)
GL_TEXCOORD4_BIT_PGI=_C('GL_TEXCOORD4_BIT_PGI',0x80000000)
GL_VERTEX23_BIT_PGI=_C('GL_VERTEX23_BIT_PGI',0x00000004)
GL_VERTEX4_BIT_PGI=_C('GL_VERTEX4_BIT_PGI',0x00000008)
GL_VERTEX_CONSISTENT_HINT_PGI=_C('GL_VERTEX_CONSISTENT_HINT_PGI',0x1A22B)
GL_VERTEX_DATA_HINT_PGI=_C('GL_VERTEX_DATA_HINT_PGI',0x1A22A)




'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.WGL import _types as _cs
# End users want this...
from OpenGL.raw.WGL._types import *
from OpenGL.raw.WGL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'WGL_ARB_create_context_robustness'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.WGL,'WGL_ARB_create_context_robustness',error_checker=_errors._error_checker)
WGL_CONTEXT_RESET_NOTIFICATION_STRATEGY_ARB=_C('WGL_CONTEXT_RESET_NOTIFICATION_STRATEGY_ARB',0x8256)
WGL_CONTEXT_ROBUST_ACCESS_BIT_ARB=_C('WGL_CONTEXT_ROBUST_ACCESS_BIT_ARB',0x00000004)
WGL_LOSE_CONTEXT_ON_RESET_ARB=_C('WGL_LOSE_CONTEXT_ON_RESET_ARB',0x8252)
WGL_NO_RESET_NOTIFICATION_ARB=_C('WGL_NO_RESET_NOTIFICATION_ARB',0x8261)




'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GL import _types as _cs
# End users want this...
from OpenGL.raw.GL._types import *
from OpenGL.raw.GL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GL_EXT_texture_swizzle'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GL,'GL_EXT_texture_swizzle',error_checker=_errors._error_checker)
GL_TEXTURE_SWIZZLE_A_EXT=_C('GL_TEXTURE_SWIZZLE_A_EXT',0x8E45)
GL_TEXTURE_SWIZZLE_B_EXT=_C('GL_TEXTURE_SWIZZLE_B_EXT',0x8E44)
GL_TEXTURE_SWIZZLE_G_EXT=_C('GL_TEXTURE_SWIZZLE_G_EXT',0x8E43)
GL_TEXTURE_SWIZZLE_RGBA_EXT=_C('GL_TEXTURE_SWIZZLE_RGBA_EXT',0x8E46)
GL_TEXTURE_SWIZZLE_R_EXT=_C('GL_TEXTURE_SWIZZLE_R_EXT',0x8E42)




'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GL import _types as _cs
# End users want this...
from OpenGL.raw.GL._types import *
from OpenGL.raw.GL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GL_SGIS_sharpen_texture'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GL,'GL_SGIS_sharpen_texture',error_checker=_errors._error_checker)
GL_LINEAR_SHARPEN_ALPHA_SGIS=_C('GL_LINEAR_SHARPEN_ALPHA_SGIS',0x80AE)
GL_LINEAR_SHARPEN_COLOR_SGIS=_C('GL_LINEAR_SHARPEN_COLOR_SGIS',0x80AF)
GL_LINEAR_SHARPEN_SGIS=_C('GL_LINEAR_SHARPEN_SGIS',0x80AD)
GL_SHARPEN_TEXTURE_FUNC_POINTS_SGIS=_C('GL_SHARPEN_TEXTURE_FUNC_POINTS_SGIS',0x80B0)
@_f
@_p.types(None,_cs.GLenum,arrays.GLfloatArray)
def glGetSharpenTexFuncSGIS(target,points):pass
@_f
@_p.types(None,_cs.GLenum,_cs.GLsizei,arrays.GLfloatArray)
def glSharpenTexFuncSGIS(target,n,points):pass



'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.EGL import _types as _cs
# End users want this...
from OpenGL.raw.EGL._types import *
from OpenGL.raw.EGL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'EGL_EXT_output_base'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.EGL,'EGL_EXT_output_base',error_checker=_errors._error_checker)
EGL_BAD_OUTPUT_LAYER_EXT=_C('EGL_BAD_OUTPUT_LAYER_EXT',0x322D)
EGL_BAD_OUTPUT_PORT_EXT=_C('EGL_BAD_OUTPUT_PORT_EXT',0x322E)
# EGL_NO_OUTPUT_LAYER_EXT=_C('EGL_NO_OUTPUT_LAYER_EXT',((EGLOutputLayerEXT)0))
# EGL_NO_OUTPUT_PORT_EXT=_C('EGL_NO_OUTPUT_PORT_EXT',((EGLOutputPortEXT)0))
EGL_SWAP_INTERVAL_EXT=_C('EGL_SWAP_INTERVAL_EXT',0x322F)
@_f
@_p.types(_cs.EGLBoolean,_cs.EGLDisplay,arrays.EGLAttribArray,ctypes.POINTER(_cs.EGLOutputLayerEXT),_cs.EGLint,arrays.GLintArray)
def eglGetOutputLayersEXT(dpy,attrib_list,layers,max_layers,num_layers):pass
@_f
@_p.types(_cs.EGLBoolean,_cs.EGLDisplay,arrays.EGLAttribArray,ctypes.POINTER(_cs.EGLOutputPortEXT),_cs.EGLint,arrays.GLintArray)
def eglGetOutputPortsEXT(dpy,attrib_list,ports,max_ports,num_ports):pass
@_f
@_p.types(_cs.EGLBoolean,_cs.EGLDisplay,_cs.EGLOutputLayerEXT,_cs.EGLint,_cs.EGLAttrib)
def eglOutputLayerAttribEXT(dpy,layer,attribute,value):pass
@_f
@_p.types(_cs.EGLBoolean,_cs.EGLDisplay,_cs.EGLOutputPortEXT,_cs.EGLint,_cs.EGLAttrib)
def eglOutputPortAttribEXT(dpy,port,attribute,value):pass
@_f
@_p.types(_cs.EGLBoolean,_cs.EGLDisplay,_cs.EGLOutputLayerEXT,_cs.EGLint,arrays.EGLAttribArray)
def eglQueryOutputLayerAttribEXT(dpy,layer,attribute,value):pass
@_f
@_p.types(ctypes.c_char_p,_cs.EGLDisplay,_cs.EGLOutputLayerEXT,_cs.EGLint)
def eglQueryOutputLayerStringEXT(dpy,layer,name):pass
@_f
@_p.types(_cs.EGLBoolean,_cs.EGLDisplay,_cs.EGLOutputPortEXT,_cs.EGLint,arrays.EGLAttribArray)
def eglQueryOutputPortAttribEXT(dpy,port,attribute,value):pass
@_f
@_p.types(ctypes.c_char_p,_cs.EGLDisplay,_cs.EGLOutputPortEXT,_cs.EGLint)
def eglQueryOutputPortStringEXT(dpy,port,name):pass



'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GL import _types as _cs
# End users want this...
from OpenGL.raw.GL._types import *
from OpenGL.raw.GL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GL_NV_sample_locations'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GL,'GL_NV_sample_locations',error_checker=_errors._error_checker)
GL_FRAMEBUFFER_PROGRAMMABLE_SAMPLE_LOCATIONS_NV=_C('GL_FRAMEBUFFER_PROGRAMMABLE_SAMPLE_LOCATIONS_NV',0x9342)
GL_FRAMEBUFFER_SAMPLE_LOCATION_PIXEL_GRID_NV=_C('GL_FRAMEBUFFER_SAMPLE_LOCATION_PIXEL_GRID_NV',0x9343)
GL_PROGRAMMABLE_SAMPLE_LOCATION_NV=_C('GL_PROGRAMMABLE_SAMPLE_LOCATION_NV',0x9341)
GL_PROGRAMMABLE_SAMPLE_LOCATION_TABLE_SIZE_NV=_C('GL_PROGRAMMABLE_SAMPLE_LOCATION_TABLE_SIZE_NV',0x9340)
GL_SAMPLE_LOCATION_NV=_C('GL_SAMPLE_LOCATION_NV',0x8E50)
GL_SAMPLE_LOCATION_PIXEL_GRID_HEIGHT_NV=_C('GL_SAMPLE_LOCATION_PIXEL_GRID_HEIGHT_NV',0x933F)
GL_SAMPLE_LOCATION_PIXEL_GRID_WIDTH_NV=_C('GL_SAMPLE_LOCATION_PIXEL_GRID_WIDTH_NV',0x933E)
GL_SAMPLE_LOCATION_SUBPIXEL_BITS_NV=_C('GL_SAMPLE_LOCATION_SUBPIXEL_BITS_NV',0x933D)
@_f
@_p.types(None,_cs.GLenum,_cs.GLuint,_cs.GLsizei,arrays.GLfloatArray)
def glFramebufferSampleLocationsfvNV(target,start,count,v):pass
@_f
@_p.types(None,_cs.GLuint,_cs.GLuint,_cs.GLsizei,arrays.GLfloatArray)
def glNamedFramebufferSampleLocationsfvNV(framebuffer,start,count,v):pass
@_f
@_p.types(None,)
def glResolveDepthValuesNV():pass



'''OpenGL extension SGIX.polynomial_ffd

This module customises the behaviour of the 
OpenGL.raw.GL.SGIX.polynomial_ffd to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/SGIX/polynomial_ffd.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.SGIX.polynomial_ffd import *
from OpenGL.raw.GL.SGIX.polynomial_ffd import _EXTENSION_NAME

def glInitPolynomialFfdSGIX():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

# INPUT glDeformationMap3dSGIX.points size not checked against 'target,ustride,uorder,vstride,vorder,wstride,worder'
glDeformationMap3dSGIX=wrapper.wrapper(glDeformationMap3dSGIX).setInputArraySize(
    'points', None
)
# INPUT glDeformationMap3fSGIX.points size not checked against 'target,ustride,uorder,vstride,vorder,wstride,worder'
glDeformationMap3fSGIX=wrapper.wrapper(glDeformationMap3fSGIX).setInputArraySize(
    'points', None
)
### END AUTOGENERATED SECTION


"""Passing of None as an array data-type
"""
REGISTRY_NAME = 'nones'
import logging
_log = logging.getLogger( 'OpenGL.arrays.nones' )

from OpenGL import acceleratesupport
NoneHandler = None
if acceleratesupport.ACCELERATE_AVAILABLE:
    try:
        from OpenGL_accelerate.nones_formathandler import NoneHandler
    except ImportError as err:
        _log.warning(
            "Unable to load nones_formathandler accelerator from OpenGL_accelerate"
        )
if NoneHandler is None:
    from OpenGL.arrays import formathandler
    class NoneHandler( formathandler.FormatHandler ):
        """Numpy-specific data-type handler for OpenGL"""
        HANDLED_TYPES = (type(None), )
        def from_param( self, value, typeCode=None  ):
            """Convert to a ctypes pointer value"""
            return None
        def dataPointer( self, value ):
            """return long for pointer value"""
            return None
        def voidDataPointer( cls, value ):
            """Given value in a known data-pointer type, return void_p for pointer"""
            return None
        def asArray( self, value, typeCode=None ):
            """Given a value, convert to array representation"""
            return None
        def arrayToGLType( self, value ):
            """Given a value, guess OpenGL type of the corresponding pointer"""
            raise TypeError( """Can't guess type of a NULL pointer""" )
        def arraySize( self, value, typeCode = None ):
            """Given a data-value, calculate dimensions for the array"""
            return 0
        def arrayByteCount( self, value, typeCode = None ):
            """Given a data-value, calculate number of bytes required to represent"""
            return 0
        def zeros( self, shape, typeCode= None ):
            """Create an array of given shape with given typeCode"""
            raise TypeError( """Can't create NULL pointer filled with values""" )
        def ones( self, shape, typeCode= None ):
            """Create an array of given shape with given typeCode"""
            raise TypeError( """Can't create NULL pointer filled with values""" )
        def unitSize( self, value, typeCode=None ):
            """Determine unit size of an array (if possible)"""
            raise TypeError( """Can't determine unit size of a null pointer""" )
        def dimensions( self, value, typeCode=None ):
            """Determine dimensions of the passed array value (if possible)"""
            return (0,)



'''OpenGL extension ARB.framebuffer_no_attachments

This module customises the behaviour of the 
OpenGL.raw.GL.ARB.framebuffer_no_attachments to provide a more 
Python-friendly API

Overview (from the spec)
	
	Framebuffer objects as introduced by ARB_framebuffer_object and OpenGL 3.0
	provide a generalized mechanism for rendering to off-screen surfaces.
	Each framebuffer object may have depth, stencil and zero or more color
	attachments that can be written to by the GL.  The size of the framebuffer
	(width, height, layer count, sample count) is derived from the attachments
	of that framebuffer.  In unextended OpenGL 4.2, it is not legal to render
	into a framebuffer object that has no attachments.  Such a framebuffer
	would be considered incomplete with the
	FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT status.
	
	With OpenGL 4.2 and ARB_shader_image_load_store, fragment shaders are
	capable of doing random access writes to buffer and texture memory via
	image loads, stores, and atomics.  This ability enables algorithms using
	the conventional rasterizer to generate a collection of fragments, where
	each fragment shader invocation will write its outputs to buffer or
	texture memory using image stores or atomics.  Such algorithms may have no
	need to write color or depth values to a conventional framebuffer.
	However, a framebuffer with no attachments will be considered incomplete
	and no rasterization or fragment shader exectuion will occur.  To avoid
	such errors, an application may be required to create an otherwise
	unnecessary "dummy" texture and attach it to the framebuffer (possibly
	with color writes masked off).  If the algorithm requires the rasterizer
	to operate over a large number of pixels, this dummy texture will
	needlessly consume a significant amount of memory.
	
	This extension enables the algorithms described above to work even with a
	framebuffer with no attachments.  Applications can specify default width,
	height, layer count, and sample count parameters for a framebuffer object.
	When a framebuffer with no attachments is bound, it will be considered
	complete as long as the application has specified non-zero default width
	and height parameters.  For the purposes of rasterization, the framebuffer
	will be considered to have a width, height, layer count, and sample count
	derived from its default parameters.  Framebuffers with one or more
	attachments are not affected by these default parameters; the size of the
	framebuffer will still be derived from the sizes of the attachments in
	that case.
	
	Additionally, this extension provides queryable implementation-dependent
	maximums for framebuffer width, height, layer count, and sample count,
	which may differ from similar limits on textures and renderbuffers.  These
	maximums will be used to error-check the default framebuffer parameters
	and also permit implementations to expose the ability to rasterize to an
	attachment-less framebuffer larger than the maximum supported texture
	size.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/ARB/framebuffer_no_attachments.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.ARB.framebuffer_no_attachments import *
from OpenGL.raw.GL.ARB.framebuffer_no_attachments import _EXTENSION_NAME

def glInitFramebufferNoAttachmentsARB():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

glGetFramebufferParameteriv=wrapper.wrapper(glGetFramebufferParameteriv).setOutput(
    'params',size=_glgets._glget_size_mapping,pnameArg='pname',orPassIn=True
)
### END AUTOGENERATED SECTION


'''OpenGL extension ATI.pn_triangles

This module customises the behaviour of the 
OpenGL.raw.GL.ATI.pn_triangles to provide a more 
Python-friendly API

Overview (from the spec)
	
	ATI_pn_triangles provides a path for enabling the GL to internally 
	tessellate input geometry into curved patches.  The extension allows the 
	user to tune the amount of tessellation to be performed on each triangle as 
	a global state value.  The intent of PN Triangle tessellation is 
	typically to produce geometry with a smoother silhouette and more organic 
	shape.
	
	The tessellated patch will replace the triangles input into the GL.  
	The GL will generate new vertices in object-space, prior to geometry 
	transformation.  Only the vertices and normals are required to produce 
	proper results, and the rest of the information per vertex is interpolated 
	linearly across the patch.  

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/ATI/pn_triangles.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.ATI.pn_triangles import *
from OpenGL.raw.GL.ATI.pn_triangles import _EXTENSION_NAME

def glInitPnTrianglesATI():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )


### END AUTOGENERATED SECTION


'''OpenGL extension EXT.draw_buffers2

This module customises the behaviour of the 
OpenGL.raw.GL.EXT.draw_buffers2 to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension builds upon the ARB_draw_buffers extension and provides
	separate blend enables and color write masks for each color output.  In
	ARB_draw_buffers (part of OpenGL 2.0), separate values can be written to
	each color buffer, but the blend enable and color write mask are global
	and apply to all color outputs.
	
	While this extension does provide separate blend enables, it does not
	provide separate blend functions or blend equations per color output.
	

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/EXT/draw_buffers2.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
import ctypes
from OpenGL.raw.GL import _types, _glgets
from OpenGL.raw.GL.EXT.draw_buffers2 import *
from OpenGL.raw.GL.EXT.draw_buffers2 import _EXTENSION_NAME

def glInitDrawBuffers2EXT():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

glGetBooleanIndexedvEXT=wrapper.wrapper(glGetBooleanIndexedvEXT).setOutput(
    'data',size=_glgets._glget_size_mapping,pnameArg='target',orPassIn=True
)
glGetIntegerIndexedvEXT=wrapper.wrapper(glGetIntegerIndexedvEXT).setOutput(
    'data',size=_glgets._glget_size_mapping,pnameArg='target',orPassIn=True
)
### END AUTOGENERATED SECTION


"""VertexBufferObject helper class

Basic usage:

    my_data = numpy.array( data, 'f')
    my_vbo = vbo.VBO( my_data )
    ...
    my_vbo.bind()
    try:
        ...
        glVertexPointer( my_vbo, ... )
        ...
        glNormalPointer( my_vbo + 12, ... )
    finally:
        my_vbo.unbind()
    
    or 
    
    with my_vbo:
        ...
        glVertexPointer( my_vbo, ... )
        ...
        glNormalPointer( my_vbo + 12, ... )        

See the OpenGLContext shader tutorials for a gentle introduction on the 
usage of VBO objects:

    http://pyopengl.sourceforge.net/context/tutorials/shader_intro.xhtml

This implementation will choose either the ARB or Core (OpenGL 1.5) 
implementation of the VBO functions.
"""
from OpenGL.arrays.arraydatatype import ArrayDatatype
from OpenGL.arrays.formathandler import FormatHandler
from OpenGL.raw.GL import _types 
from OpenGL import error
from OpenGL._bytes import bytes,unicode,as_8_bit
import ctypes,logging
_log = logging.getLogger( 'OpenGL.arrays.vbo' )
from OpenGL._bytes import long, integer_types

import weakref
__all__ = ('VBO','VBOHandler','mapVBO')

class Implementation( object ):
    """Abstraction point for the various implementations that can be used
    """
    IMPLEMENTATION_CLASSES = []
    CHOSEN = None
    @classmethod
    def register( cls ):
        cls.IMPLEMENTATION_CLASSES.append( cls )
    
    @classmethod 
    def get_implementation( cls, *args ):
        if cls.CHOSEN is None:
            for possible in cls.IMPLEMENTATION_CLASSES:
                implementation = possible()
                if possible:
                    Implementation.CHOSEN = implementation
                    break
        return cls.CHOSEN

    EXPORTED_NAMES = '''glGenBuffers
    glBindBuffer 
    glBufferData 
    glBufferSubData 
    glDeleteBuffers
    glMapBuffer
    glUnmapBuffer
    GL_STATIC_DRAW
    GL_STATIC_READ
    GL_STATIC_COPY
    GL_DYNAMIC_DRAW
    GL_DYNAMIC_READ
    GL_DYNAMIC_COPY
    GL_STREAM_DRAW
    GL_STREAM_READ
    GL_STREAM_COPY
    GL_ARRAY_BUFFER
    GL_ELEMENT_ARRAY_BUFFER
    GL_UNIFORM_BUFFER
    GL_TEXTURE_BUFFER
    GL_TRANSFORM_FEEDBACK_BUFFER'''.split()
    available = False
    def _arbname( self, name ):
        return (
            (name.startswith( 'gl' ) and name.endswith( 'ARB' )) or
            (name.startswith( 'GL_' ) and name.endswith( 'ARB' ))
        ) and (name != 'glInitVertexBufferObjectARB')
    def basename( self, name ):
        if name.endswith( '_ARB' ):
            return name[:-4]
        elif name.endswith( 'ARB' ):
            return name[:-3]
        else:
            return name
    def __nonzero__( self ):
        return self.available
    __bool__ = __nonzero__
    def deleter( self, buffers, key):
        """Produce a deleter callback to delete the given buffer"""
        # these values are stored here to avoid them being cleaned up 
        # to non during module deletion and causing errors to be raised
        nfe = error.NullFunctionError
        gluint = _types.GLuint
        def doBufferDeletion( *args, **named ):
            while buffers:
                try:
                    buffer = buffers.pop()
                except IndexError as err:
                    break
                else:
                    try:
                        # Note that to avoid ERROR_ON_COPY issues
                        # we have to pass an array-compatible type here...
                        buf = gluint( buffer )
                        self.glDeleteBuffers(1, buf)
                    except (AttributeError, nfe) as err:
                        pass
            try:
                self._DELETERS_.pop( key )
            except KeyError as err:
                pass
        return doBufferDeletion
    _DELETERS_ = {}

get_implementation = Implementation.get_implementation

from OpenGL import acceleratesupport
VBO = None
if acceleratesupport.ACCELERATE_AVAILABLE:
    try:
        from OpenGL_accelerate.vbo import (
            VBO,VBOOffset,VBOHandler,VBOOffsetHandler,
        )
    except ImportError as err:
        _log.warning(
            "Unable to load VBO accelerator from OpenGL_accelerate"
        )
if VBO is None:
    class VBO( object ):
        """Instances can be passed into array-handling routines

        You can check for whether VBOs are supported by accessing the implementation:

            if bool(vbo.get_implementation()):
                # vbo version of code
            else:
                # fallback version of code
        """
        copied = False
        _no_cache_ = True # do not cache in context data arrays
        def __init__(
            self, data, usage='GL_DYNAMIC_DRAW',
            target='GL_ARRAY_BUFFER', size=None,
        ):
            """Initialize the VBO object 
            
            data -- PyOpenGL-compatible array-data structure, numpy arrays, ctypes arrays, etc.
            usage -- OpenGL usage constant describing expected data-flow patterns (this is a hint 
                to the GL about where/how to cache the data)
                
                GL_STATIC_DRAW_ARB
                GL_STATIC_READ_ARB
                GL_STATIC_COPY_ARB
                GL_DYNAMIC_DRAW_ARB
                GL_DYNAMIC_READ_ARB
                GL_DYNAMIC_COPY_ARB
                GL_STREAM_DRAW_ARB
                GL_STREAM_READ_ARB
                GL_STREAM_COPY_ARB
                
                DRAW constants suggest to the card that the data will be primarily used to draw 
                on the card.  READ that the data will be read back into the GL.  COPY means that 
                the data will be used both for DRAW and READ operations.
                
                STATIC suggests that the data will only be written once (or a small number of times).
                DYNAMIC suggests that the data will be used a small number of times before being 
                discarded.
                STREAM suggests that the data will be updated approximately every time that it is 
                used (that is, it will likely only be used once).
                
            target -- VBO target to which to bind (array or indices)
                GL_ARRAY_BUFFER -- array-data binding 
                GL_ELEMENT_ARRAY_BUFFER -- index-data binding
                GL_UNIFORM_BUFFER -- used to pass mid-size arrays of data packed into a buffer
                GL_TEXTURE_BUFFER -- used to pass large arrays of data as a pseudo-texture
                GL_TRANSFORM_FEEDBACK_BUFFER -- used to receive transformed vertices for processing
                
            size -- if not provided, will use arrayByteCount to determine the size of the data-array,
                thus this value (number of bytes) is required when using opaque data-structures,
                (such as ctypes pointers) as the array data-source.
            """
            self.usage = usage
            self.set_array( data, size )
            self.target = target
            self.buffers = []
            self._copy_segments = []
        _I_ = None
        implementation = property( get_implementation, )
        def resolve( self, value ):
            """Resolve string constant to constant"""
            if isinstance( value, (bytes,unicode)):
                return getattr( self.implementation, self.implementation.basename( value ) )
            return value
        def set_array( self, data, size=None ):
            """Update our entire array with new data
            
            data -- PyOpenGL-compatible array-data structure, numpy arrays, ctypes arrays, etc.
            size -- if not provided, will use arrayByteCount to determine the size of the data-array,
                thus this value (number of bytes) is required when using opaque data-structures,
                (such as ctypes pointers) as the array data-source.
            """
            self.data = data
            self.copied = False
            if size is not None:
                self.size = size
            elif self.data is not None:
                self.size = ArrayDatatype.arrayByteCount( self.data )
        def __setitem__( self, slice, array):
            """Set slice of data on the array and vbo (if copied already)

            slice -- the Python slice object determining how the data should
                be copied into the vbo/array
            array -- something array-compatible that will be used as the
                source of the data, note that the data-format will have to
                be the same as the internal data-array to work properly, if
                not, the amount of data copied will be wrong.

            This is a reasonably complex operation, it has to have all sorts
            of state-aware changes to correctly map the source into the low-level
            OpenGL view of the buffer (which is just bytes as far as the GL
            is concerned).
            """
            if slice.step and not slice.step == 1:
                raise NotImplemented( """Don't know how to map stepped arrays yet""" )
            # TODO: handle e.g. mapping character data into an integer data-set
            data = ArrayDatatype.asArray( array )
            data_length = ArrayDatatype.arrayByteCount( array )
            start = (slice.start or 0)
            stop = (slice.stop or len(self.data))
            if start < 0:
                start += len(self.data)
                start = max((start,0))
            if stop < 0:
                stop += len(self.data)
                stop = max((stop,0))
            self.data[ slice ] = data
            if self.copied and self.buffers:
                if start-stop == len(self.data):
                    # re-copy the whole data-set
                    self.copied = False
                elif len(data):
                    # now the fun part, we need to make the array match the
                    # structure of the array we're going to copy into and make
                    # the "size" parameter match the value we're going to copy in,
                    # note that a 2D array (rather than a 1D array) may require
                    # multiple mappings to copy into the memory area...

                    # find the step size from the dimensions and base size...
                    size = ArrayDatatype.arrayByteCount( self.data[0] )
                    #baseSize = ArrayDatatype.unitSize( data )
                    # now create the start and distance values...
                    start *= size
                    stop *= size
                    # wait until the last moment (bind) to copy the data...
                    self._copy_segments.append(
                        (start,(stop-start), data)
                    )
        def __len__( self ):
            """Delegate length/truth checks to our data-array"""
            return len( self.data )
        def __getattr__( self, key ):
            """Delegate failing attribute lookups to our data-array"""
            if key not in ('data','usage','target','buffers', 'copied','_I_','implementation','_copy_segments' ):
                return getattr( self.data, key )
            else:
                raise AttributeError( key )
        def create_buffers( self ):
            """Create the internal buffer(s)"""
            assert not self.buffers, """Already created the buffer"""
            self.buffers = [ long(self.implementation.glGenBuffers(1)) ]
            self.target = self.resolve( self.target )
            self.usage = self.resolve( self.usage )
            self.implementation._DELETERS_[ id(self) ] = weakref.ref( self, self.implementation.deleter( self.buffers, id(self) ))
            return self.buffers
        def copy_data( self ):
            """Copy our data into the buffer on the GL side (if required)
            
            Ensures that the GL's version of the data in the VBO matches our 
            internal view of the data, either by copying the entire data-set 
            over with glBufferData or by updating the already-transferred 
            data with glBufferSubData.
            """
            assert self.buffers, """Should do create_buffers before copy_data"""
            if self.copied:
                if self._copy_segments:
                    while self._copy_segments:
                        start,size,data  = self._copy_segments.pop(0)
                        dataptr = ArrayDatatype.voidDataPointer( data )
                        self.implementation.glBufferSubData(self.target, start, size, dataptr)
            else:
                if self.data is not None and self.size is None:
                    self.size = ArrayDatatype.arrayByteCount( self.data )
                self.implementation.glBufferData(
                    self.target,
                    self.size,
                    self.data,
                    self.usage,
                )
                self.copied = True
        def delete( self ):
            """Delete this buffer explicitly"""
            if self.buffers:
                while self.buffers:
                    try:
                        self.implementation.glDeleteBuffers(1, self.buffers.pop(0))
                    except (AttributeError,error.NullFunctionError) as err:
                        pass
        def __int__( self ):
            """Get our VBO id"""
            if not self.buffers:
                self.create_buffers()
            return self.buffers[0]
        def bind( self ):
            """Bind this buffer for use in vertex calls
            
            If we have not yet created our implementation-level VBO, then we 
            will create it before binding.  Once bound, calls self.copy_data()
            """
            if not self.buffers:
                buffers = self.create_buffers()
            self.implementation.glBindBuffer( self.target, self.buffers[0])
            self.copy_data()
        def unbind( self ):
            """Unbind the buffer (make normal array operations active)"""
            self.implementation.glBindBuffer( self.target,0 )

        def __add__( self, other ):
            """Add an integer to this VBO (create a VBOOffset)"""
            if hasattr( other, 'offset' ):
                other = other.offset
            assert isinstance( other, integer_types ), """Only know how to add integer/long offsets"""
            return VBOOffset( self, other )

        __enter__ = bind
        def __exit__( self, exc_type=None, exc_val=None, exc_tb=None ):
            """Context manager exit"""
            self.unbind()
            return False # do not supress exceptions...

    class VBOOffset( object ):
        """Offset into a VBO instance 
        
        This class is normally instantiated by doing a my_vbo + int operation,
        it can be passed to VBO requiring operations and will generate the 
        appropriate integer offset value to be passed in.
        """
        def __init__( self, vbo, offset ):
            """Initialize the offset with vbo and offset (unsigned integer)"""
            self.vbo = vbo
            self.offset = offset
        def __getattr__( self, key ):
            """Delegate any undefined attribute save vbo to our vbo"""
            if key != 'vbo':
                return getattr( self.vbo, key )
            raise AttributeError( 'No %r key in VBOOffset'%(key,))
        def __add__( self, other ):
            """Allow adding integers or other VBOOffset instances 
            
            returns a VBOOffset to the this VBO with other.offset + self.offset
            or, if other has no offset, returns VBOOffset with self.offset + other
            """
            if hasattr( other, 'offset' ):
                other = other.offset
            return VBOOffset( self.vbo, self.offset + other )

    class VBOHandler( FormatHandler ):
        """Handles VBO instances passed in as array data
        
        This FormatHandler is registered with PyOpenGL on import of this module 
        to provide handling of VBO objects as array data-sources
        """
        vp0 = ctypes.c_void_p( 0 )
        def dataPointer( self, instance ):
            """Retrieve data-pointer from the instance's data

            Is always NULL, to indicate use of the bound pointer
            """
            return 0
        def from_param( self, instance, typeCode=None ):
            """Always returns c_void_p(0)"""
            return self.vp0
        def zeros( self, dims, typeCode ):
            """Not implemented"""
            raise NotImplemented( """Don't have VBO output support yet""" )
        ones = zeros
        def asArray( self, value, typeCode=None ):
            """Given a value, convert to array representation"""
            return value
        def arrayToGLType( self, value ):
            """Given a value, guess OpenGL type of the corresponding pointer"""
            return ArrayDatatype.arrayToGLType( value.data )
        def arrayByteCount( self, value ):
            return ArrayDatatype.arrayByteCount( value.data )
        def arraySize( self, value, typeCode = None ):
            """Given a data-value, calculate dimensions for the array"""
            return ArrayDatatype.arraySize( value.data )
        def unitSize( self, value, typeCode=None ):
            """Determine unit size of an array (if possible)"""
            return ArrayDatatype.unitSize( value.data )
        def dimensions( self, value, typeCode=None ):
            """Determine dimensions of the passed array value (if possible)"""
            return ArrayDatatype.dimensions( value.data )

    class VBOOffsetHandler( VBOHandler ):
        """Handles VBOOffset instances passed in as array data
        
        Registered on module import to provide support for VBOOffset instances 
        as sources for array data.
        """
        def dataPointer( self, instance ):
            """Retrieve data-pointer from the instance's data

            returns instance' offset
            """
            return instance.offset
        def from_param( self, instance, typeCode=None ):
            """Returns a c_void_p( instance.offset )"""
            return ctypes.c_void_p( instance.offset )

_cleaners = {}
def _cleaner( vbo ):
    """Construct a mapped-array cleaner function to unmap vbo.target"""
    def clean( ref ):
        try:
            _cleaners.pop( vbo )
        except Exception as err:
            pass
        else:
            vbo.implementation.glUnmapBuffer( vbo.target )
    return clean

def mapVBO( vbo, access=0x88BA ): # GL_READ_WRITE
    """Map the given buffer into a numpy array...

    Method taken from:
     http://www.mail-archive.com/numpy-discussion@lists.sourceforge.net/msg01161.html

    This should be considered an *experimental* API,
    it is not guaranteed to be available in future revisions
    of this library!
    
    Simplification to use ctypes cast from comment by 'sashimi' on my blog...
    """
    from numpy import frombuffer
    vp = vbo.implementation.glMapBuffer( vbo.target, access )
    # TODO: obviously this is not the right way to do this should allow each format 
    # handler to convert the pointer in their own way...
    vp_array = ctypes.cast(vp, ctypes.POINTER(ctypes.c_byte*vbo.size) )
    # Note: we could have returned the raw ctypes.c_byte array instead...
    array = frombuffer( vp_array, 'B' )
    _cleaners[vbo] = weakref.ref( array, _cleaner( vbo ))
    return array
